#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" The initial context factory for Python file system contexts. """


# Local imports.
from context import Context
from initial_context_factory import InitialContextFactory
from object_serializer import ObjectSerializer
from pyfs_context import PyFSContext
from pyfs_context_factory import PyFSContextFactory
from pyfs_object_factory import PyFSObjectFactory
from pyfs_state_factory import PyFSStateFactory


class PyFSInitialContextFactory(InitialContextFactory):
    """ The initial context factory for Python file system contexts. """

    ###########################################################################
    # 'InitialContextFactory' interface.
    ###########################################################################

    def get_initial_context(self, environment):
        """ Creates an initial context for beginning name resolution. """

        # Object factories.
        object_factories = [PyFSObjectFactory(), PyFSContextFactory()]
        environment[Context.OBJECT_FACTORIES] = object_factories

        # State factories.
        state_factories = [PyFSStateFactory()]
        environment[Context.STATE_FACTORIES] = state_factories

        # Object serializers.
        object_serializers = [ObjectSerializer()]
        environment[PyFSContext.OBJECT_SERIALIZERS] = object_serializers

        return PyFSContext(path=r'', environment=environment)

#### EOF ######################################################################



""" Context adapter factory for trait dicts. """


# Enthought library imports.
from apptools.naming.api import ContextAdapterFactory
from traits.api import Str, TraitDict

# Local imports.
from trait_dict_context_adapter import TraitDictContextAdapter


class TraitDictContextAdapterFactory(ContextAdapterFactory):
    """ Context adapter factoryfor Python trait dicts. """

    #### 'ContextAdapterFactory' interface ####################################

    # The type of object that we adapt.
    adaptee_class = TraitDict

    #### 'TraitDictContextAdapterFactory' interface ###########################

    # The name of the trait (on the adaptee) that provides the trait dict.
    trait_name = Str

    ###########################################################################
    # Protected 'AbstractAdapterFactory' interface.
    ###########################################################################

    def _adapt(self, adaptee, target_class, environment, context):
        """ Returns an adapter that adapts an object to the target class. """

        adapter = TraitDictContextAdapter(
            adaptee     = adaptee,
            environment = environment,
            context     = context,
            trait_name  = self.trait_name
        )

        return adapter

#### EOF ######################################################################



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#------------------------------------------------------------------------------

""" Defines the various color editors for the Wx user interface toolkit.
"""

#------------------------------------------------------------------------------
#  Imports:
#------------------------------------------------------------------------------

import wx
import wx.combo

from traits.api import List, TraitError
from traitsui.editors.color_editor \
    import ToolkitEditorFactory as BaseToolkitEditorFactory

from editor_factory import SimpleEditor as BaseSimpleEditor
from editor_factory import ReadonlyEditor as BaseReadonlyEditor
from editor_factory import TextEditor as BaseTextEditor

from color_trait import w3c_color_database
from helper import TraitsUIPanel

# Version dependent imports (ColourPtr not defined in wxPython 2.5):
try:
    ColorTypes = (wx.Colour, wx.ColourPtr)
except:
    ColorTypes = wx.Colour


#---------------------------------------------------------------------------
#  The Wx ToolkitEditorFactory class.
#---------------------------------------------------------------------------

class ToolkitEditorFactory(BaseToolkitEditorFactory):
    """ Wx editor factory for color editors.
    """

    def to_wx_color(self, editor, color=None):
        """ Gets the wxPython color equivalent of the object trait.
        """
        if color is None:
            if self.mapped:
                color = getattr(editor.object, editor.name + '_')
            else:
                color = getattr(editor.object, editor.name)

        if isinstance(color, tuple):
            color = wx.Colour(*[int(round(c * 255.0)) for c in color])
        return color

    #--------------------------------------------------------------------------
    #  Gets the application equivalent of a wxPython value:
    #--------------------------------------------------------------------------

    def from_wx_color(self, color):
        """ Gets the application equivalent of a wxPython value.
        """
        return color.Red(), color.Green(), color.Blue()

    #--------------------------------------------------------------------------
    #  Returns the text representation of a specified color value:
    #--------------------------------------------------------------------------

    def str_color(self, color):
        """ Returns the text representation of a specified color value.
        """
        if isinstance(color, ColorTypes):
            alpha = color.Alpha()
            if alpha == 255:
                return "rgb(%d,%d,%d)" % (
                    color.Red(), color.Green(), color.Blue())

            return "rgb(%d,%d,%d,%d)" % (
                color.Red(), color.Green(), color.Blue(), alpha)

        return str(color)


#------------------------------------------------------------------------------
#  'ColorComboBox' class:
#------------------------------------------------------------------------------

class ColorComboBox(wx.combo.OwnerDrawnComboBox):

    def OnDrawItem(self, dc, rect, item, flags):

        r = wx.Rect(rect.x, rect.y, rect.width, rect.height)
        r.Deflate(3, 0)
        swatch_size = r.height - 2

        color_name = self.GetString(item)

        dc.DrawText(color_name, r.x + 3,
                    r.y + (r.height - dc.GetCharHeight()) / 2)

        if color_name == 'custom':
            swatch = wx.Rect(r.x + r.width - swatch_size, r.y + 1,
                             swatch_size, swatch_size)
            dc.GradientFillLinear(swatch, wx.Colour(255, 255, 0),
                                  wx.Colour(0, 0, 255))
        else:
            color = w3c_color_database.Find(color_name)

            brush = wx.Brush(color)
            dc.SetBrush(brush)
            dc.DrawRectangle(r.x + r.width - swatch_size, r.y + 1,
                             swatch_size, swatch_size)


#------------------------------------------------------------------------------
#  'SimpleColorEditor' class:
#------------------------------------------------------------------------------

class SimpleColorEditor(BaseSimpleEditor):
    """ Simple style of color editor, which displays a text field whose
    background color is the color value. Selecting the text field displays
    a dialog box for selecting a new color value.
    """

    #--------------------------------------------------------------------------
    #  Invokes the pop-up editor for an object trait:
    #--------------------------------------------------------------------------

    choices = List()

    def _choices_default(self):
        """ by default, uses the W3C 16 color names.
        """
        return ['aqua', 'black', 'blue', 'fuchsia', 'gray', 'green',
                'lime', 'maroon', 'navy', 'olive', 'purple', 'red',
                'silver', 'teal', 'white', 'yellow', 'custom']

    def init(self, parent):
        """
        Finishes initializing the editor by creating the underlying widget.
        """
        current_color = self.factory.to_wx_color(self)
        current_color_name = current_color.GetAsString()

        self.control = ColorComboBox(parent, -
                                     1, current_color_name, wx.Point(0, 0), wx.Size(40, -
                                                                                    1), self.choices, style=wx.wx.CB_READONLY)

        self.control.Bind(wx.EVT_COMBOBOX, self.color_selected)
        return

    #--------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #--------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        current_color = self.factory.to_wx_color(self)
        self.control.SetValue(self.string_value(current_color))

    def color_selected(self, event):
        """
        Event for when color is selected
        """

        color_name = self.choices[event.Selection]

        if color_name == 'custom':
            color_dialog = wx.ColourDialog(self.control)
            result = color_dialog.ShowModal()
            if result == wx.ID_CANCEL:
                return

            color = color_dialog.GetColourData().GetColour()
            self.value = self.factory.from_wx_color(color)
        else:
            try:
                color = w3c_color_database.Find(color_name)
                self.value = self.factory.from_wx_color(color)
            except ValueError:
                pass

        return

    def string_value(self, color):
        """ Returns the text representation of a specified color value.
        """
        color_name = w3c_color_database.FindName(color)
        if color_name != '':
            return color_name

        return color.GetAsString()


#------------------------------------------------------------------------------
#  'CustomColorEditor' class:
#------------------------------------------------------------------------------

class CustomColorEditor(BaseSimpleEditor):
    """ Simple style of color editor, which displays a text field whose
    background color is the color value. Selecting the text field displays
    a dialog box for selecting a new color value.
    """

    #--------------------------------------------------------------------------
    #  Invokes the pop-up editor for an object trait:
    #--------------------------------------------------------------------------

    def init(self, parent):
        """
        Finishes initializing the editor by creating the underlying widget.
        """
        self.control = self._panel = parent = TraitsUIPanel(parent, -1)
        sizer = wx.BoxSizer(wx.HORIZONTAL)

        # 'text_control' is the text display of the color.
        text_control = wx.TextCtrl(parent, -1, self.str_value,
                                   style=wx.TE_PROCESS_ENTER)
        wx.EVT_KILL_FOCUS(text_control, self.update_object)
        wx.EVT_TEXT_ENTER(parent, text_control.GetId(), self.update_object)

        # 'button_control' shows the 'Edit' button.
        button_control = wx.Button(parent, label='Edit', style=wx.BU_EXACTFIT)
        wx.EVT_BUTTON(button_control, button_control.GetId(),
                      self.open_color_dialog)

        sizer.Add(text_control, wx.ALIGN_LEFT)
        sizer.AddSpacer(8)
        sizer.Add(button_control, wx.ALIGN_RIGHT)
        self.control.SetSizer(sizer)

        self._text_control = text_control
        self._button_control = button_control

        self.set_tooltip()

        return

    def update_object(self, event):
        """ Handles the user changing the contents of the edit control.
        """
        if not isinstance(event, wx._core.CommandEvent):
            event.Skip()
            return
        try:
            # The TextCtrl object was saved as self._text_control in init().
            value = self._text_control.GetValue()
            self.value = w3c_color_database.Find(value)
            self.set_color()
        except TraitError:
            pass

    def set_color(self):
        # The CustomColorEditor uses this method instead of the global
        # set_color function.
        color = self.factory.to_wx_color(self)
        self._text_control.SetBackgroundColour(color)
        self.control.SetBackgroundColour(color)
        self._text_control.SetValue(self.string_value(color))

    #--------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #--------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        self.set_color()

    def open_color_dialog(self, event):
        """ Opens the color dialog and sets the value upon return
        """
        color_data = wx.ColourData()
        color_data.SetColour(self.value)

        color_dialog = wx.ColourDialog(self.control, color_data)
        result = color_dialog.ShowModal()
        if result == wx.ID_CANCEL:
            return

        color = color_dialog.GetColourData().GetColour()
        self.value = color
        self.set_color()

    def color_selected(self, event):
        """
        Event for when color is selected
        """
        color = event.GetColour()
        try:
            self.value = self.factory.from_wx_color(color)
        except ValueError:
            pass

        return

    def string_value(self, color):
        """ Returns the text representation of a specified color value.
        """
        color_name = w3c_color_database.FindName(color)
        if color_name != '':
            return color_name

        return self.factory.str_color(color)


#------------------------------------------------------------------------------
#  'ReadonlyColorEditor' class:
#------------------------------------------------------------------------------

class ReadonlyColorEditor(BaseReadonlyEditor):
    """ Read-only style of color editor, which displays a read-only text field
    whose background color is the color value.
    """

    #--------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #--------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self.control = wx.TextCtrl(parent, style=wx.TE_READONLY)

    #--------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #--------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        #super(ReadonlyColorEditor, self).update_editor()
        self.control.SetValue(self.string_value(self.value))
        set_color(self)

    #--------------------------------------------------------------------------
    #  Returns the text representation of a specified color value:
    #--------------------------------------------------------------------------

    def string_value(self, color):
        """ Returns the text representation of a specified color value.
        """
        color_name = w3c_color_database.FindName(color)
        if color_name != '':
            return color_name

        return self.factory.str_color(color)


#------------------------------------------------------------------------------
#  'ReadonlyColorEditor' class:
#------------------------------------------------------------------------------

class TextColorEditor(BaseTextEditor):
    """ Text style of color editor, which displays a text field
    whose background color is the color value.
    """

    #--------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #--------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        self.control.SetValue(self.string_value(self.value))
        set_color(self)

    def update_object(self, event):
        """ Handles the user changing the contents of the edit control.
        """
        if not isinstance(event, wx._core.CommandEvent):
            return
        try:
            self.value = w3c_color_database.Find(self.control.GetValue())
            set_color(self)
        except TraitError:
            pass

    #--------------------------------------------------------------------------
    #  Returns the text representation of a specified color value:
    #--------------------------------------------------------------------------

    def string_value(self, color):
        """ Returns the text representation of a specified color value.
        """
        color_name = w3c_color_database.FindName(color)
        if color_name != '':
            return color_name

        return self.factory.str_color(color)


#------------------------------------------------------------------------------
#   Sets the color of the specified editor's color control:
#------------------------------------------------------------------------------

def set_color(editor):
    """  Sets the color of the specified color control.
    """
    color = editor.factory.to_wx_color(editor)
    editor.control.SetBackgroundColour(color)


# Define the SimpleEditor, CustomEditor, etc. classes which are used by the
# editor factory for the color editor.
SimpleEditor = SimpleColorEditor
CustomEditor = CustomColorEditor
TextEditor = TextColorEditor
ReadonlyEditor = ReadonlyColorEditor



import os
import sys
import wx.lib.inspection

import fs

# Enthought library imports.
from pyface.api import ImageResource, FileDialog, YES, NO, OK, CANCEL
from pyface.action.api import Action, ActionItem, Group
from pyface.tasks.action.api import EditorAction
from traits.api import on_trait_change, Property, Instance, Bool, Str, Unicode, Any, List, Int

from omnivore.framework.about import AboutDialog
from omnivore.utils.file_guess import FileGuess
from omnivore.utils.wx.dialogs import get_file_dialog_wildcard
from omnivore.utils.wx.error_logger import show_logging_frame
from omnivore.templates import iter_templates

import logging
log = logging.getLogger(__name__)


class DocumentationOnlyAction(Action):
    ""
    enabled = False
    description = ""
    tooltip = ""


class ShowLoggerAction(Action):
    """ Displays a window that can be used to turn on debugging of particular
    parts of the program.

    The log levels shown initially are the default log levels for each logger.
    Using the ``Filter`` text entry box, you can enter a string or a comma
    separated list of strings that will be used to select which loggers get
    switched to *DEBUG* mode. Everything else gets set to its default state,
    usually either *INFO* or *WARNING*.

    The string is not a regular expression, but will match partial strings.
    """
    name = "Show Debug Log"
    tooltip = "Open a window to view and manage debug logging"

    def perform(self, event=None):
        show_logging_frame()


class NewFileAction(Action):
    """A list of supported file types that can be created by Omnivore.

    Choosing an item in this list will open up a new tab in the current window
    showing a new blank template of the selected file format.

    Available templates include:
    """
    doc_hint = "parent,list"
    tooltip = Property(Unicode, depends_on='name')

    task_id = Any

    uri = Str

    def perform(self, event=None):
        task = event.task.window.application.find_or_create_task_of_type(self.task_id)
        log.debug("Loading %s as %s" % (self.uri, task))
        guess = FileGuess(self.uri)
        task.new(guess)

    def _get_tooltip(self):
        return u'Open a new %s' % self.name


class NewFileGroup(Group):
    """ A menu for creating a new file for each type of task
    """

    #### 'ActionManager' interface ############################################

    id = 'NewFileGroup'

    items = List

    #### 'TaskChangeMenuManager' interface ####################################

    # The ActionManager to which the group belongs.
    manager = Any

    # The window that contains the group.
    application = Instance('envisage.ui.tasks.api.TasksApplication')

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_items(self):
        items = []

        for template in iter_templates():
            name = template["label"]
            task_id = self.application.find_best_task_id(template["task"])
            log.debug("NewFileAction for %s as %s" % (name, task_id))
            action = NewFileAction(name=name, uri=template["uri"], task_id=task_id)
            items.append((name, ActionItem(action=action)))
        items.sort()
        items = [i[1] for i in items]
        return items

    def _rebuild(self):
        # Clear out the old group, then build the new one.
        self.destroy()
        self.items = self._get_items()

        # Inform our manager that it needs to be rebuilt.
        self.manager.changed = True

    #### Trait initializers ###################################################

    def _items_default(self):
        self.application.on_trait_change(self._rebuild, 'task_factories[]')
        return self._get_items()

    def _manager_default(self):
        manager = self
        while isinstance(manager, Group):
            manager = manager.parent
        return manager

    def _application_default(self):
        return self.manager.controller.task.window.application


class OpenAction(Action):
    """Open a file using a file select dialog box.

    """
    name = 'Open...'
    accelerator = 'Ctrl+O'
    tooltip = 'Open a file'
    image = ImageResource('file_open')

    def perform(self, event):
        path = event.task.prompt_local_file_dialog()
        if path is not None:
            event.task.window.application.load_file(path, event.task)


class SaveAction(EditorAction):
    """Save the file, overwriting the previously saved version

    """
    name = 'Save'
    accelerator = 'Ctrl+S'
    tooltip = 'Save the current file'
    image = ImageResource('file_save')
    enabled_name = 'can_save' # enabled based on state of task.active_editor.can_save

    def perform(self, event):
        self.active_editor.save(None)


class SaveAsAction(EditorAction):
    """Save the file to a new filename, leaving the originally loaded unchanged
    on disk.

    """
    name = 'Save As...'
    accelerator = 'Ctrl+Shift+S'
    tooltip = 'Save the current file with a new name'
    image = ImageResource('file_save_as')

    def perform(self, event):
        path = event.task.prompt_local_file_dialog(save=True, default_filename=self.active_editor.document.name)
        if path:
            self.active_editor.save(path, saver=self.active_editor.encode_data)


class RevertAction(EditorAction):
    """Reverts the file to the last saved version on disk

    This throws away any edits and is not undoable.
    """
    name = 'Revert'
    tooltip = 'Revert to last saved version'

    def perform(self, event):
        uri = self.active_editor.document.metadata.uri
        message = "Revert file from\n\n%s?" % uri
        result = event.task.confirm(message=message, title='Revert File?')
        if result:
            try:
                guess = FileGuess(uri)
                document = event.task.window.application.guess_document(guess)
                self.active_editor.load(document)
            except fs.errors.FSError, e:
                event.task.error("Can't revert from %s:\n\n%s" % (uri, str(e)), 'Revert Error')


class PageSetupAction(Action):
    name = 'Page Setup...'
    tooltip = 'Choose options for printing'

    def perform(self, event):
        event.task.page_setup()


class PrintPreviewAction(EditorAction):
    name = 'Print Preview'
    tooltip = 'Preview the pages to be printed'
    enabled_name = 'printable'

    def perform(self, event):
        self.active_editor.print_preview()


class PrintAction(EditorAction):
    name = 'Print...'
    tooltip = 'Print the current view to a printer'
    enabled_name = 'printable'

    def perform(self, event):
        self.active_editor.print_page()


class SaveAsPDFAction(EditorAction):
    name = 'Export As PDF...'
    tooltip = 'Save the current view as a PDF (if possible with this editor)'
    enabled_name = 'printable'

    def perform(self, event):
        path = event.task.prompt_local_file_dialog(save=True, title="Save PDF")
        if path:
            self.active_editor.save_as_pdf(path)


class SaveAsImageAction(EditorAction):
    name = 'Save As Image...'
    tooltip = 'Save the current view as an image (if possible with this editor)'
    enabled_name = 'imageable'

    def perform(self, event):
        path = event.task.prompt_local_file_dialog(save=True, title="Save Image")
        if path:
            self.active_editor.save_as_image(path)


class ExitAction(Action):
    name = 'Quit'
    accelerator = 'Ctrl+Q'
    tooltip = 'Quit the program'
    menu_role = "Quit"

    def perform(self, event):
        event.task.window.application.exit()


class NameChangeAction(EditorAction):
    """Extension to the EditorAction that provides a user-updatable menu item
    name based on a trait
    
    EditorAction is subclassed from ListeningAction, and the ListeningAction
    methods destroy and _object_chaged must be called because the new
    trait name 'menu_item_name' can't be added to list of traits managed by
    ListeningAction and so must be taken care of here before the superclass
    can do its work.
    """
    menu_item_name = Str

    def destroy(self):
        """ Called when the action is no longer required.

        Remove all the task listeners.

        """

        if self.object:
            self.object.on_trait_change(
                self._menu_item_update, self.menu_item_name, remove=True
            )
        super(NameChangeAction, self).destroy()

    def _menu_item_name_changed(self, old, new):
        obj = self.object
        if obj is not None:
            if old:
                obj.on_trait_change(self._menu_item_update, old, remove=True)
            if new:
                obj.on_trait_change(self._menu_item_update, new)
        self._label_update()

    def _object_changed(self, old, new):
        kind = 'menu_item'
        method = getattr(self, '_%s_update' % kind)
        name = getattr(self, '%s_name' % kind)
        if name:
            if old:
                old.on_trait_change(method, name, remove=True)
            if new:
                new.on_trait_change(method, name)
        method()
        super(NameChangeAction, self)._object_changed(old, new)

    def _menu_item_update(self):
        if self.menu_item_name:
            if self.object:
                self.name = str(self._get_attr(self.object,
                                               self.menu_item_name, 'Undo'))
            else:
                self.name = getattr(self, 'default_name', 'Undo')
        else:
            self.name = getattr(self, 'default_name', 'Undo')


class UndoAction(NameChangeAction):
    """Undo the last action

    Actions that modify data are undoable; some that modify the metadata are
    but movement commands are not stored in the undo list, so for example
    cursor moves or changes to selection regions are not undoable.
    """
    default_name = 'Undo'
    name = default_name
    accelerator = 'Ctrl+Z'
    tooltip = 'Undo last action'
    image = ImageResource('undo')
    enabled_name = 'can_undo'
    menu_item_name = 'undo_label'

    def perform(self, event):
        self.active_editor.undo()


class RedoAction(NameChangeAction):
    """Redo the last operation that was undone.  See `Undo`_.
    """
    default_name = 'Redo'
    name = default_name
    accelerator = 'Ctrl+Shift+Z'
    tooltip = 'Redo the last undone action'
    image = ImageResource('redo')
    enabled_name = 'can_redo'
    menu_item_name = 'redo_label'

    def perform(self, event):
        self.active_editor.redo()


class CutAction(EditorAction):
    name = 'Cut'
    accelerator = 'Ctrl+X'
    tooltip = 'Cut and remove the current selection'
    enabled_name = 'can_cut'

    def perform(self, event):
        self.active_editor.cut()


class CopyAction(EditorAction):
    name = 'Copy'
    accelerator = 'Ctrl+C'
    tooltip = 'Copy the current selection'
    enabled_name = 'can_copy'

    def perform(self, event):
        self.active_editor.copy()


class PasteAction(EditorAction):
    name = 'Paste'
    accelerator = 'Ctrl+V'
    tooltip = 'Paste from the clipboard'
    enabled_name = 'can_paste'

    def perform(self, event):
        self.active_editor.paste()


class SelectAllAction(Action):
    name = 'Select All'
    accelerator = 'Ctrl+A'
    tooltip = 'Select the entire document'

    def perform(self, event):
        event.task.active_editor.select_all()


class SelectNoneAction(Action):
    name = 'Select None'
    accelerator = 'Shift+Ctrl+A'
    tooltip = 'Clear selection'

    def perform(self, event):
        event.task.active_editor.select_none()


class SelectInvertAction(Action):
    """Inverts the selection; that is, select everything that is currently unselected and unselect those that were selected.

    """
    name = 'Invert Selection'
    tooltip = 'Invert selection'

    def perform(self, event):
        event.task.active_editor.select_invert()


class PreferencesAction(Action):
    """Open a window to change program settings and defaults.

    """
    name = 'Preferences...'
    accelerator = 'Ctrl+,'
    tooltip = 'Program settings and configuration options'
    menu_role = "Preferences"

    def perform(self, event):
        # FIXME: using the built-in dialog box handler segfaults on OS X:
        #0   libwx_osx_cocoau-2.9.4.0.0.dylib	0x0000000101539f07 wxWidgetCocoaImpl::mouseEvent(NSEvent*, NSView*, void*) + 183
        #1   com.apple.AppKit              	0x00007fff89efac98 -[NSWindow sendEvent:] + 6306
        #2   libwx_osx_cocoau-2.9.4.0.0.dylib	0x0000000101526a5c -[wxNSPanel sendEvent:] + 140
        #3   com.apple.AppKit              	0x00007fff89e943a5 -[NSApplication sendEvent:] + 5593
        #4   com.apple.AppKit              	0x00007fff8a0e2797 -[NSApplication _realDoModalLoop:peek:] + 708
        #5   com.apple.AppKit              	0x00007fff8a0e2369 -[NSApplication runModalForWindow:] + 120
        #6   libwx_osx_cocoau-2.9.4.0.0.dylib	0x000000010151804e wxModalEventLoop::DoRun() + 174
        #7   libwx_osx_cocoau-2.9.4.0.0.dylib	0x00000001013eb607 wxCFEventLoop::Run() + 55
        #8   libwx_osx_cocoau-2.9.4.0.0.dylib	0x00000001014538d9 wxDialog::ShowModal() + 73
        #9   _windows_.so                  	0x000000010422b7b6 _wrap_Dialog_ShowModal + 102
        if False:
            from envisage.ui.tasks.preferences_dialog import \
                PreferencesDialog

            window = event.task.window
            dialog = window.application.get_service(PreferencesDialog)
            ui = dialog.edit_traits(parent=window.control, kind='livemodal')

            if ui.result:
                window.application.preferences.save()

        from omnivore.framework.preferences import FrameworkPreferenceDialog
        dialog = FrameworkPreferenceDialog(application=event.task.window.application, style="modal")
        status = dialog.open()
        if status == OK:
            event.task.window.application.preferences_changed_event = True


class AboutAction(Action):
    name = 'About...'
    tooltip = 'Display window with version number and author info'
    menu_role = "About"

    def perform(self, event):
        # Don't rely on the event window as this might be called by the OSX
        # minimal menu which, once installed by the OSX plugin, never changes
        # with the task.
        top = event.task.window.application.active_window
        AboutDialog(top.control, top.active_task)


class OpenLogDirectoryAction(Action):
    """Open the log directory in the desktop file manager program.

    The log directory will contain debug logs (if enabled) and other files,
    most of which are generally only useful for developers or to get more
    information to send to the developers in the event of a problem.
    """
    name = 'Open Log Directory in File Manager'
    tooltip = 'Open the log directory in the desktop file manager program'

    def perform(self, event):
        app = event.task.window.application
        filename = app.get_log_file_name("dummy")
        dirname = os.path.dirname(filename)
        import subprocess
        if sys.platform.startswith("win"):
            file_manager = 'explorer'
        elif sys.platform == "darwin":
            file_manager = '/usr/bin/open'
        else:
            file_manager = 'xdg-open'
        subprocess.call([file_manager, dirname])


class UserGuideAction(Action):
    name = "User Guide"
    tooltip = 'Display the user guide in a new window'

    def perform(self, event):
        # Don't rely on the event window as this might be called by the OSX
        # minimal menu which, once installed by the OSX plugin, never changes
        # with the task.
        event.task.window.application.show_help()


class NewWindowAction(Action):
    """Open a new, blank Omnivore window that can be used to load new files
    or to provide a second view on a currently loaded file.

    """
    name = 'New Window'
    tooltip = 'Open a new window'

    def perform(self, event):
        event.task.new_window()


class WidgetInspectorAction(Action):
    name = 'Widget Inspector'
    tooltip = 'Open the wxPython Widget Inspector'

    def perform(self, event):
        wx.lib.inspection.InspectionTool().Show()


class BaseDynamicSubmenuGroup(Group):
    """ A group used for a dynamic menu.
    """

    #### 'ActionManager' interface ############################################

    id = 'DynamicMenuGroup'
    items = List

    #### 'TaskChangeMenuManager' interface ####################################

    # The ActionManager to which the group belongs.
    manager = Any

    # ENTHOUGHT QUIRK: This doesn't work: can't have a property depending on
    # a task because this forces task_default to be called very early in the
    # initialization process, before the window hierarchy is defined.
    #
    # active_editor = Property(Instance(IEditor),
    #                         depends_on='task.active_editor')

    event_name = Str('change this to the Event trait')

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_items(self, event_data=None):
        # Override this in your subclass to return the list of actions
        return []

    def _rebuild(self, new_trait_val):
        # Clear out the old group, then build the new one.
        self.destroy()

        # Get the new items, passing the event arguments to the method
        self.items = self._get_items(new_trait_val)

        # Set up parent so that radio items can determine their siblings to
        # uncheck others when checked. (see the _checked_changed method in
        # pyface/ui/wx/action/action_item.py)
        for item in self.items:
            item.parent = self

        # Inform our manager that it needs to be rebuilt.
        self.manager.changed = True

    #### Trait initializers ###################################################

    def _items_default(self):
        log.debug("DYNAMICGROUP: _items_default!!!")
        t = self._get_trait_for_event()
        t.on_trait_change(self._rebuild, self.event_name)
        return self._get_items()

    def _get_trait_for_event(self):
        raise NotImplementedError

    def _manager_default(self):
        manager = self
        while isinstance(manager, Group):
            manager = manager.parent
        log.debug("DYNAMICGROUP: _manager_default=%s!!!" % manager)
        return manager

    # ENTHOUGHT QUIRK: This doesn't work: the trait change decorator never
    # seems to get called, however specifying the on_trait_change in the
    # _items_default method works.
    #
    #    @on_trait_change('task.layer_selection_changed')
    #    def updated_fired(self, event):
    #        log.debug("SAVELAYERGROUP: updated!!!")
    #        self._rebuild(event)


class TaskDynamicSubmenuGroup(BaseDynamicSubmenuGroup):
    """ A group used for a dynamic menu.
    """

    # The task instance must be passed in as an attribute creation argument
    # because we need to bind on a task trait change to update the menu
    task = Instance('omnivore.framework.task.FrameworkTask')

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_trait_for_event(self):
        return self.task

    #### Trait initializers ###################################################

    def _task_default(self):
        log.debug("DYNAMICGROUP: _task_default=%s!!!" % self.manager.controller.task)
        return self.manager.controller.task


class ApplicationDynamicSubmenuGroup(BaseDynamicSubmenuGroup):
    """ A group used for a dynamic menu based on an application event.
    """

    # The application instance must be a trait so we can set an on_trait_change
    # handler
    application = Instance('envisage.ui.tasks.api.TasksApplication')

    def _get_trait_for_event(self):
        return self.application

    #### Trait initializers ###################################################

    def _application_default(self):
        log.debug("DYNAMICGROUP: _application_default=%s!!!" % self.manager.controller.task.window.application)
        return self.manager.controller.task.window.application


class SwitchDocumentAction(Action):
    doc_hint = "ignore"

    document_id = Int

    name = Str

    def perform(self, event):
        app = event.task.window.application
        doc = app.get_document(self.document_id)
        if doc is not None:
            log.debug("Switching to %s, task=%s" % (doc, doc.last_task_id))
            task = app.find_active_task_of_type(doc.last_task_id)
            task.find_tab_or_open(doc)


class DocumentSelectGroup(ApplicationDynamicSubmenuGroup):
    """ A menu for creating a new file for each type of task
    """

    event_name = 'successfully_loaded_event'

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_items(self, event_data=None):
        documents = sorted(self.application.documents, key=lambda x: x.name.lower())
        items = []
        for document in documents:
            action = SwitchDocumentAction(document_id=document.document_id, name=document.menu_name)
            items.append(ActionItem(action=action))
        return items


class NewViewInNewTaskAction(Action):
    factory_id = Str

    def perform(self, event):
        event.task.window.application.create_task_from_factory_id(event.task.active_editor, self.factor_id)


class NewViewInGroup(TaskDynamicSubmenuGroup):
    name = '<this is not used; the name is set in the SMenu in which this group is added>'
    tooltip = 'New view of the project with a different task'
    event_name = 'document_changed'

    def perform(self, event):
        event.task.new_window(view=event.task.active_editor)

    @on_trait_change('task.document_changed')
    def _update_name(self):
        if self.task.active_editor:
            # The name of the menu is taken from the SMenu in which this Group
            # is included, not the name in this object.
            self.manager.name = "New View of %s" % self.task.active_editor.document.name
            # Tell the MenuManager in the hierarchy that the SMenu name has
            # changed.  This depth up the hierarchy (grandparent) works for
            # this instance where this group is a child of the SMen.
            self.manager.parent.parent.changed = True

    def _get_items(self, event_data=None):
        e = self.task.active_editor
        items = []
        if e:
            factories = self.task.window.application.get_possible_task_factories(e.document)
            for factory in factories:
                action = NewViewInNewTaskAction(name="In a %s Tab" % factory.name, factor_id=factory.id)
                items.append(ActionItem(action=action))
        return items



"""
fs.wrapfs.subfs
===============

An FS wrapper class for accessing just a subdirectory for an FS.

"""

from fs.wrapfs import WrapFS
from fs.errors import *
from fs.path import *


class SubFS(WrapFS):
    """A SubFS represents a sub directory of another filesystem object.

    SubFS objects are returned by opendir, which effectively creates a
    'sandbox' filesystem that can only access files/dirs under a root path
    within its 'parent' dir.
    """

    def __init__(self, wrapped_fs, sub_dir):
        self.sub_dir = abspath(normpath(sub_dir))
        super(SubFS, self).__init__(wrapped_fs)

    def _encode(self, path):
        return pathjoin(self.sub_dir, relpath(normpath(path)))

    def _decode(self, path):
        return abspath(normpath(path))[len(self.sub_dir):]

    def __str__(self):
        #return self.wrapped_fs.desc(self.sub_dir)
        return '<SubFS: %s/%s>' % (self.wrapped_fs, self.sub_dir.lstrip('/'))

    def __unicode__(self):
        return u'<SubFS: %s/%s>' % (self.wrapped_fs, self.sub_dir.lstrip('/'))

    def __repr__(self):
        return "SubFS(%r, %r)" % (self.wrapped_fs, self.sub_dir)

    def desc(self, path):
        if path in ('', '/'):
            return self.wrapped_fs.desc(self.sub_dir)
        return '%s!%s' % (self.wrapped_fs.desc(self.sub_dir), path)

    def setcontents(self, path, data, encoding=None, errors=None, chunk_size=64*1024):
        path = self._encode(path)
        return self.wrapped_fs.setcontents(path, data, chunk_size=chunk_size)

    def opendir(self, path):
        if not self.exists(path):
            raise ResourceNotFoundError(path)
        path = self._encode(path)
        return self.wrapped_fs.opendir(path)

    def close(self):
        self.closed = True

    def removedir(self, path, recursive=False, force=False):
        # Careful not to recurse outside the subdir
        path = normpath(path)
        if path in ('', '/'):
            raise RemoveRootError(path)
        super(SubFS, self).removedir(path, force=force)
        if recursive:
            try:
                if dirname(path) not in ('', '/'):
                    self.removedir(dirname(path), recursive=True)
            except DirectoryNotEmptyError:
                pass

#        if path in ("","/"):
#            if not force:
#                for path2 in self.listdir(path):
#                    raise DirectoryNotEmptyError(path)
#            else:
#                for path2 in self.listdir(path,absolute=True,files_only=True):
#                    try:
#                        self.remove(path2)
#                    except ResourceNotFoundError:
#                        pass
#                for path2 in self.listdir(path,absolute=True,dirs_only=True):
#                    try:
#                        self.removedir(path2,force=True)
#                    except ResourceNotFoundError:
#                        pass
#        else:
#            super(SubFS,self).removedir(path,force=force)
#            if recursive:
#                try:
#                    if dirname(path):
#                        self.removedir(dirname(path),recursive=True)
#                except DirectoryNotEmptyError:
#                    pass





#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/07/2004
#
#------------------------------------------------------------------------------

""" Defines the abstract Editor class, which represents an editing control for
    an object trait in a Traits-based user interface.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import (
    Any,
    Bool,
    HasPrivateTraits,
    HasTraits,
    Instance,
    Property,
    ReadOnly,
    Str,
    Trait,
    TraitError,
    TraitListEvent,
    Undefined,
    cached_property)

from traits.trait_base import not_none

from .editor_factory import EditorFactory

from .context_value import ContextValue

from .undo import UndoItem

from .item import Item

#-------------------------------------------------------------------------
#  Trait definitions:
#-------------------------------------------------------------------------

# Reference to an EditorFactory object
factory_trait = Trait(EditorFactory)

#-------------------------------------------------------------------------
#  'Editor' abstract base class:
#-------------------------------------------------------------------------


class Editor(HasPrivateTraits):
    """ Represents an editing control for an object trait in a Traits-based
        user interface.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # The UI (user interface) this editor is part of:
    ui = Instance('traitsui.ui.UI')

    # Full name of the object the editor is editing (e.g.
    # 'object.link1.link2'):
    object_name = Str('object')

    # The object this editor is editing (e.g. object.link1.link2):
    object = Instance(HasTraits)

    # The name of the trait this editor is editing (e.g. 'value'):
    name = ReadOnly

    # The context object the editor is editing (e.g. object):
    context_object = Property

    # The extended name of the object trait being edited. That is,
    # 'object_name.name' minus the context object name at the beginning. For
    # example: 'link1.link2.value':
    extended_name = Property

    # Original value of object.name (e.g. object.link1.link2.value):
    old_value = Any

    # Text description of the object trait being edited:
    description = ReadOnly

    # The Item object used to create this editor:
    item = Instance(Item, ())

    # The GUI widget defined by this editor:
    control = Any

    # The GUI label (if any) defined by this editor:
    label_control = Any

    # Is the underlying GUI widget enabled?
    enabled = Bool(True)

    # Is the underlying GUI widget visible?
    visible = Bool(True)

    # Is the underlying GUI widget scrollable?
    scrollable = Bool(False)

    # The EditorFactory used to create this editor:
    factory = factory_trait

    # Is the editor updating the object.name value?
    updating = Bool(False)

    # Current value for object.name:
    value = Property

    # Current value of object trait as a string:
    str_value = Property

    # The trait the editor is editing (not its value, but the trait itself):
    value_trait = Property

    # The current editor invalid state status:
    invalid = Bool(False)

    #-------------------------------------------------------------------------
    #  Initializes the object:
    #-------------------------------------------------------------------------

    def __init__(self, parent, **traits):
        """ Initializes the editor object.
        """
        HasPrivateTraits.__init__(self, **traits)
        try:
            self.old_value = getattr(self.object, self.name)
        except AttributeError:
            ctrait = self.object.base_trait(self.name)
            if ctrait.type == 'event' or self.name == 'spring':
                # Getting the attribute will fail for 'Event' traits:
                self.old_value = Undefined
            else:
                raise

        # Synchronize the application invalid state status with the editor's:
        self.sync_value(self.factory.invalid, 'invalid', 'from')

    #-------------------------------------------------------------------------
    #  Finishes editor set-up:
    #-------------------------------------------------------------------------

    def prepare(self, parent):
        """ Finishes setting up the editor.
        """
        name = self.extended_name
        if name != 'None':
            self.context_object.on_trait_change(self._update_editor, name,
                                                dispatch='ui')
        self.init(parent)
        self._sync_values()
        self.update_editor()

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        raise NotImplementedError

    #-------------------------------------------------------------------------
    #  Assigns focus to the editor's underlying toolkit widget:
    #-------------------------------------------------------------------------

    def set_focus(self):
        """ Assigns focus to the editor's underlying toolkit widget.
        """
        raise NotImplementedError

    #-------------------------------------------------------------------------
    #  Disposes of the contents of an editor:
    #-------------------------------------------------------------------------

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        if self.ui is None:
            return

        name = self.extended_name
        if name != 'None':
            self.context_object.on_trait_change(self._update_editor, name,
                                                remove=True)

        if self._user_from is not None:
            for name, handler in self._user_from:
                self.on_trait_change(handler, name, remove=True)

        if self._user_to is not None:
            for object, name, handler in self._user_to:
                object.on_trait_change(handler, name, remove=True)

        # Break linkages to references we no longer need:
        self.object = self.ui = self.item = self.factory = self.control = \
            self.label_control = self.old_value = self._context_object = None

    #-------------------------------------------------------------------------
    #  Returns the context object the editor is using (Property implementation):
    #-------------------------------------------------------------------------

    @cached_property
    def _get_context_object(self):
        """ Returns the context object the editor is using (Property
            implementation).
        """
        object_name = self.object_name
        context_key = object_name.split('.', 1)[0]
        if (object_name != '') and (context_key in self.ui.context):
            return self.ui.context[context_key]

        # This handles the case of a 'ListItemProxy', which is not in the
        # ui.context, but is the editor 'object':
        return self.object

    #-------------------------------------------------------------------------
    #  Returns the extended trait name being edited (Property implementation):
    #-------------------------------------------------------------------------

    @cached_property
    def _get_extended_name(self):
        """ Returns the extended trait name being edited (Property
            implementation).
        """
        return ('%s.%s' % (self.object_name, self.name)).split('.', 1)[1]

    #-------------------------------------------------------------------------
    #  Returns the trait the editor is editing (Property implementation):
    #-------------------------------------------------------------------------

    def _get_value_trait(self):
        """ Returns the trait the editor is editing (Property implementation).
        """
        return self.object.trait(self.name)

    #-------------------------------------------------------------------------
    #  Gets/Sets the associated object trait's value:
    #-------------------------------------------------------------------------

    def _get_value(self):
        return getattr(self.object, self.name, Undefined)

    def _set_value(self, value):
        if self.ui and self.name != 'None':
            self.ui.do_undoable(self.__set_value, value)

    def __set_value(self, value):
        self._no_update = True
        try:
            try:
                handler = self.ui.handler
                obj_name = self.object_name
                name = self.name
                method = (getattr(handler, '%s_%s_setattr' % (obj_name,
                                                              name), None) or
                          getattr(handler, '%s_setattr' % name, None) or
                          getattr(handler, 'setattr'))
                method(self.ui.info, self.object, name, value)
            except TraitError as excp:
                self.error(excp)
                raise
        finally:
            self._no_update = False

    #-------------------------------------------------------------------------
    #  Returns the text representation of a specified object trait value:
    #-------------------------------------------------------------------------

    def string_value(self, value, format_func=None):
        """ Returns the text representation of a specified object trait value.

            If the **format_func** attribute is set on the editor factory, then
            this method calls that function to do the formatting.  If the
            **format_str** attribute is set on the editor factory, then this
            method uses that string for formatting. If neither attribute is
            set, then this method just calls the built-in unicode() function.
        """
        factory = self.factory
        if factory.format_func is not None:
            return factory.format_func(value)

        if factory.format_str != '':
            return factory.format_str % value

        if format_func is not None:
            return format_func(value)

        return unicode(value)

    #-------------------------------------------------------------------------
    #  Returns the text representation of the object trait:
    #-------------------------------------------------------------------------

    def _get_str_value(self):
        """ Returns the text representation of the object trait.
        """
        return self.string_value(getattr(self.object, self.name, Undefined))

    #-------------------------------------------------------------------------
    #  Returns the text representation of a specified value:
    #-------------------------------------------------------------------------

    def _str(self, value):
        """ Returns the text representation of a specified value.
        """
        # In Unicode!
        return unicode(value)

    #-------------------------------------------------------------------------
    #  Handles an error that occurs while setting the object's trait value:
    #
    #  (Should normally be overridden in a subclass)
    #-------------------------------------------------------------------------

    def error(self, excp):
        """ Handles an error that occurs while setting the object's trait value.
        """
        pass

    #-------------------------------------------------------------------------
    #  Performs updates when the object trait changes:
    #-------------------------------------------------------------------------

    def _update_editor(self, object, name, old_value, new_value):
        """ Performs updates when the object trait changes.
        """
        # If background threads have modified the trait the editor is bound to,
        # their trait notifications are queued to the UI thread. It is possible
        # that by the time the UI thread dispatches these events, the UI the
        # editor is part of has already been closed. So we need to check if we
        # are still bound to a live UI, and if not, exit immediately:
        if self.ui is None:
            return

        # If the notification is for an object different than the one actually
        # being edited, it is due to editing an item of the form:
        # object.link1.link2.name, where one of the 'link' objects may have
        # been modified. In this case, we need to rebind the current object
        # being edited:
        if object is not self.object:
            self.object = eval(self.object_name, globals(), self.ui.context)

        # If the editor has gone away for some reason, disconnect and exit:
        if self.control is None:
            self.context_object.on_trait_change(
                self._update_editor, self.extended_name, remove=True)
            return

        # Log the change that was made (as long as it is not for an event):
        if object.base_trait(name).type != 'event':
            self.log_change(self.get_undo_item, object, name,
                            old_value, new_value)

        # If the change was not caused by the editor itself:
        if not self._no_update:
            # Update the editor control to reflect the current object state:
            self.update_editor()

    #-------------------------------------------------------------------------
    #  Logs a change made in the editor:
    #-------------------------------------------------------------------------

    def log_change(self, undo_factory, *undo_args):
        """ Logs a change made in the editor.
        """
        # Indicate that the contents of the user interface have been changed:
        ui = self.ui
        ui.modified = True

        # Create an undo history entry if we are maintaining a history:
        undoable = ui._undoable
        if undoable >= 0:
            history = ui.history
            if history is not None:
                item = undo_factory(*undo_args)
                if item is not None:
                    if undoable == history.now:
                        # Create a new undo transaction:
                        history.add(item)
                    else:
                        # Extend the most recent undo transaction:
                        history.extend(item)

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #
    #  (Should normally be overridden in a subclass)
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        pass

    #-------------------------------------------------------------------------
    #  Creates an undo history entry:
    #
    #  (Can be overridden in a subclass for special value types)
    #-------------------------------------------------------------------------

    def get_undo_item(self, object, name, old_value, new_value):
        """ Creates an undo history entry.
        """
        return UndoItem(object=object,
                        name=name,
                        old_value=old_value,
                        new_value=new_value)

    #-------------------------------------------------------------------------
    #  Returns a tuple of the form ( context_object, name[.name...], callable )
    #  for a specified extended name of the form: name or
    #  context_object_name.name[.name...]:
    #-------------------------------------------------------------------------

    def parse_extended_name(self, name):
        """ Returns a tuple of the form ( context_object, 'name[.name...],
            callable ) for a specified extended name of the form: 'name' or
            'context_object_name.name[.name...]'.
        """
        col = name.find('.')
        if col < 0:
            object = self.context_object
        else:
            object, name = self.ui.context[name[: col]], name[col + 1:]

        return (object, name, eval("lambda obj=object: obj." + name))

    #-------------------------------------------------------------------------
    #  Initializes and synchronizes (as needed) editor traits with the value of
    #  corresponding factory traits:
    #-------------------------------------------------------------------------

    def _sync_values(self):
        """ Initializes and synchronizes (as needed) editor traits with the
            value of corresponding factory traits.
        """
        factory = self.factory
        for name, trait in factory.traits(sync_value=not_none):
            value = getattr(factory, name)
            if isinstance(value, ContextValue):
                self_trait = self.trait(name)
                self.sync_value(value.name, name,
                                self_trait.sync_value or trait.sync_value,
                                self_trait.is_list is True)
            elif value is not Undefined:
                setattr(self, name, value)

    #-------------------------------------------------------------------------
    #  Sets/Unsets synchronization between an editor trait and a user object
    #  trait:
    #-------------------------------------------------------------------------

    def sync_value(self, user_name, editor_name, mode='both',
                   is_list=False):
        """ Sets or unsets synchronization between an editor trait and a user
            object trait.
        """
        if user_name != '':
            key = '%s:%s' % (user_name, editor_name)

            if self._no_trait_update is None:
                self._no_trait_update = {}

            user_ref = 'user_object'
            col = user_name.find('.')
            if col < 0:
                user_object = self.context_object
                xuser_name = user_name
            else:
                user_object = self.ui.context[user_name[: col]]
                user_name = xuser_name = user_name[col + 1:]
                col = user_name.rfind('.')
                if col >= 0:
                    user_ref += ('.' + user_name[: col])
                    user_name = user_name[col + 1:]

            user_value = compile('%s.%s' % (user_ref, user_name),
                                 '<string>', 'eval')
            user_ref = compile(user_ref, '<string>', 'eval')

            if mode in ('from', 'both'):

                def user_trait_modified(new):
                    # Need this to include 'user_object' in closure:
                    user_object
                    if key not in self._no_trait_update:
                        self._no_trait_update[key] = None
                        try:
                            setattr(self, editor_name, new)
                        except:
                            from traitsui.api import raise_to_debug
                            raise_to_debug()
                        del self._no_trait_update[key]

                user_object.on_trait_change(user_trait_modified, xuser_name)

                if self._user_to is None:
                    self._user_to = []
                self._user_to.append((user_object, xuser_name,
                                      user_trait_modified))

                if is_list:

                    def user_list_modified(event):
                        if isinstance(event, TraitListEvent):
                            if key not in self._no_trait_update:
                                self._no_trait_update[key] = None
                                n = event.index
                                try:
                                    getattr(self, editor_name)[
                                        n: n + len(event.removed)] = event.added
                                except:
                                    from traitsui.api import raise_to_debug
                                    raise_to_debug()
                                del self._no_trait_update[key]

                    user_object.on_trait_change(user_list_modified,
                                                xuser_name + '_items')
                    self._user_to.append((user_object, xuser_name + '_items',
                                          user_list_modified))

                try:
                    setattr(self, editor_name, eval(user_value))
                except:
                    from traitsui.api import raise_to_debug
                    raise_to_debug()

            if mode in ('to', 'both'):

                def editor_trait_modified(new):
                    # Need this to include 'user_object' in closure:
                    user_object
                    if key not in self._no_trait_update:
                        self._no_trait_update[key] = None
                        try:
                            setattr(eval(user_ref), user_name, new)
                        except:
                            from traitsui.api import raise_to_debug
                            raise_to_debug()
                        del self._no_trait_update[key]

                self.on_trait_change(editor_trait_modified, editor_name)

                if self._user_from is None:
                    self._user_from = []
                self._user_from.append((editor_name, editor_trait_modified))

                if is_list:

                    def editor_list_modified(event):
                        # Need this to include 'user_object' in closure:
                        user_object
                        if key not in self._no_trait_update:
                            self._no_trait_update[key] = None
                            n = event.index
                            try:
                                eval(user_value)[
                                    n: n + len(event.removed)] = event.added
                            except:
                                from traitsui.api import raise_to_debug
                                raise_to_debug()
                            del self._no_trait_update[key]

                    self.on_trait_change(editor_list_modified,
                                         editor_name + '_items')
                    self._user_from.append((editor_name + '_items',
                                            editor_list_modified))

                if mode == 'to':
                    try:
                        setattr(eval(user_ref), user_name,
                                getattr(self, editor_name))
                    except:
                        from traitsui.api import raise_to_debug
                        raise_to_debug()

    #-- UI preference save/restore interface ---------------------------------

    #-------------------------------------------------------------------------
    #  Restores any saved user preference information associated with the
    #  editor:
    #-------------------------------------------------------------------------

    def restore_prefs(self, prefs):
        """ Restores any saved user preference information associated with the
            editor.
        """
        pass

    #-------------------------------------------------------------------------
    #  Returns any user preference information associated with the editor:
    #-------------------------------------------------------------------------

    def save_prefs(self):
        """ Returns any user preference information associated with the editor.
        """
        return None



""" An editor manager that uses contributed editors. """

# Enthought library imports.
from pyface.workbench.api import EditorManager, TraitsUIEditor


class WorkbenchEditorManager(EditorManager):
    """ An editor manager that uses contributed editors. """

    ###########################################################################
    # 'IEditorManager' interface.
    ###########################################################################

    def create_editor(self, window, obj, kind):
        """ Create an editor for an object.

        For now, the 'kind' is actually a factory that produces editors. It
        should be a callable with the following signature::

            callable(window=window, obj=obj) -> IEditor
        """

        if kind is None:
            kind = TraitsUIEditor

        editor = kind(window=window, obj=obj)

        self.add_editor(editor, kind)

        return editor

    ###########################################################################
    # 'Protected' 'EditorManager'  interface.
    ###########################################################################

    def _is_editing(self, editor, obj, kind):
        """ Return True if the editor is editing the object. """

        if kind is None:
            kind = TraitsUIEditor
        return self.get_editor_kind(editor) is kind and editor.obj == obj

#### EOF ######################################################################



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   07/19/2005
#
#------------------------------------------------------------------------------

""" Defines the wxPython implementation of the editor used to wrap a non-Traits
based custom control.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

# FIXME: ToolkitEditorFactory is a proxy class defined here just for backward
# compatibility. The class has been moved to the
# traitsui.editors.custom_editor file.
from traitsui.editors.custom_editor \
    import ToolkitEditorFactory

from editor \
    import Editor

from helper \
    import open_fbi

#-------------------------------------------------------------------------
#  'CustomEditor' class:
#-------------------------------------------------------------------------


class CustomEditor(Editor):
    """ Wrapper for a custom editor control
    """
    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        factory = self.factory.factory
        if factory is not None:
            try:
                self.control = factory(*((parent, self) + self.factory.args))
            except:
                open_fbi()
        if self.control is None:
            self.control = control = wx.StaticText(
                parent, -1, 'An error occurred creating a custom editor.\n'
                'Please contact the developer.')
            control.SetBackgroundColour(wx.RED)
            control.SetForegroundColour(wx.WHITE)
        self.set_tooltip()

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        pass

### EOF #######################################################################



#-------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   06/06/2007
#
#-------------------------------------------------------------------------

""" Class to aid in automatically computing the 'slice' points for a specified
    ImageResource and then drawing it that it can be 'stretched' to fit a larger
    region than the original image.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

from colorsys \
    import rgb_to_hls

from numpy \
    import reshape, fromstring, uint8

from traits.api \
    import HasPrivateTraits, Instance, Int, List, Color, Enum, Bool

from pyface.image_resource \
    import ImageResource

from constants \
    import WindowColor

from constants import is_mac
import traitsui.wx.constants

#-------------------------------------------------------------------------
#  Recursively paint the parent's background if they have an associated image
#  slice.
#-------------------------------------------------------------------------


def paint_parent(dc, window):
    """ Recursively paint the parent's background if they have an associated
        image slice.
    """
    parent = window.GetParent()
    slice = getattr(parent, '_image_slice', None)
    if slice is not None:
        x, y = window.GetPositionTuple()
        dx, dy = parent.GetSizeTuple()
        slice.fill(dc, -x, -y, dx, dy)
    else:
        # Otherwise, just paint the normal window background color:
        dx, dy = window.GetClientSizeTuple()
        if is_mac and hasattr(window, '_border') and window._border:
            dc.SetBackgroundMode(wx.TRANSPARENT)
            dc.SetBrush(wx.Brush(wx.Colour(0, 0, 0, 0)))
        else:
            dc.SetBrush(wx.Brush(parent.GetBackgroundColour()))
        dc.SetPen(wx.TRANSPARENT_PEN)
        dc.DrawRectangle(0, 0, dx, dy)

    return slice

#-------------------------------------------------------------------------
#  'ImageSlice' class:
#-------------------------------------------------------------------------


class ImageSlice(HasPrivateTraits):

    #-- Trait Definitions ----------------------------------------------------

    # The ImageResource to be sliced and drawn:
    image = Instance(ImageResource)

    # The minimum number of adjacent, identical rows/columns needed to identify
    # a repeatable section:
    threshold = Int(10)

    # The maximum number of 'stretchable' rows and columns:
    stretch_rows = Enum(1, 2)
    stretch_columns = Enum(1, 2)

    # Width/height of the image borders:
    top = Int
    bottom = Int
    left = Int
    right = Int

    # Width/height of the extended image borders:
    xtop = Int
    xbottom = Int
    xleft = Int
    xright = Int

    # The color to use for content text:
    content_color = Instance(wx.Colour)

    # The color to use for label text:
    label_color = Instance(wx.Colour)

    # The background color of the image:
    bg_color = Color

    # Should debugging slice lines be drawn?
    debug = Bool(False)

    #-- Private Traits -------------------------------------------------------

    # The current image's opaque bitmap:
    opaque_bitmap = Instance(wx.Bitmap)

    # The current image's transparent bitmap:
    transparent_bitmap = Instance(wx.Bitmap)

    # Size of the current image:
    dx = Int
    dy = Int

    # Size of the current image's slices:
    dxs = List
    dys = List

    # Fixed minimum size of current image:
    fdx = Int
    fdy = Int

    #-- Public Methods -------------------------------------------------------

    def fill(self, dc, x, y, dx, dy, transparent=False):
        """ 'Stretch fill' the specified region of a device context with the
            sliced image.
        """
        # Create the source image dc:
        idc = wx.MemoryDC()
        if transparent:
            idc.SelectObject(self.transparent_bitmap)
        else:
            idc.SelectObject(self.opaque_bitmap)

        # Set up the drawing parameters:
        sdx, sdy = self.dx, self.dx
        dxs, dys = self.dxs, self.dys
        tdx, tdy = dx - self.fdx, dy - self.fdy

        # Calculate vertical slice sizes to use for source and destination:
        n = len(dxs)
        if n == 1:
            pdxs = [(0, 0), (1, max(1, tdx / 2)), (sdx - 2, sdx - 2),
                    (1, max(1, tdx - (tdx / 2))), (0, 0)]
        elif n == 3:
            pdxs = [(dxs[0], dxs[0]), (dxs[1], max(0, tdx)), (0, 0),
                    (0, 0), (dxs[2], dxs[2])]
        else:
            pdxs = [(dxs[0], dxs[0]), (dxs[1], max(0, tdx / 2)),
                    (dxs[2], dxs[2]), (dxs[3], max(0, tdx - (tdx / 2))),
                    (dxs[4], dxs[4])]

        # Calculate horizontal slice sizes to use for source and destination:
        n = len(dys)
        if n == 1:
            pdys = [(0, 0), (1, max(1, tdy / 2)), (sdy - 2, sdy - 2),
                    (1, max(1, tdy - (tdy / 2))), (0, 0)]
        elif n == 3:
            pdys = [(dys[0], dys[0]), (dys[1], max(0, tdy)), (0, 0),
                    (0, 0), (dys[2], dys[2])]
        else:
            pdys = [(dys[0], dys[0]), (dys[1], max(0, tdy / 2)),
                    (dys[2], dys[2]), (dys[3], max(0, tdy - (tdy / 2))),
                    (dys[4], dys[4])]

        # Iterate over each cell, performing a stretch fill from the source
        # image to the destination window:
        last_x, last_y = x + dx, y + dy
        y0, iy0 = y, 0
        for idy, wdy in pdys:
            if y0 >= last_y:
                break

            if wdy != 0:
                x0, ix0 = x, 0
                for idx, wdx in pdxs:
                    if x0 >= last_x:
                        break

                    if wdx != 0:
                        self._fill(idc, ix0, iy0, idx, idy,
                                   dc, x0, y0, wdx, wdy)
                        x0 += wdx
                    ix0 += idx
                y0 += wdy
            iy0 += idy

        if self.debug:
            dc.SetPen(wx.Pen(wx.RED))
            dc.DrawLine(x, y + self.top, last_x, y + self.top)
            dc.DrawLine(x, last_y - self.bottom - 1,
                        last_x, last_y - self.bottom - 1)
            dc.DrawLine(x + self.left, y, x + self.left, last_y)
            dc.DrawLine(last_x - self.right - 1, y,
                        last_x - self.right - 1, last_y)

    #-- Event Handlers -------------------------------------------------------

    def _image_changed(self, image):
        """ Handles the 'image' trait being changed.
        """
        # Save the original bitmap as the transparent version:
        self.transparent_bitmap = bitmap = \
            image.create_image().ConvertToBitmap()

        # Save the bitmap size information:
        self.dx = dx = bitmap.GetWidth()
        self.dy = dy = bitmap.GetHeight()

        # Create the opaque version of the bitmap:
        self.opaque_bitmap = wx.EmptyBitmap(dx, dy)
        mdc2 = wx.MemoryDC()
        mdc2.SelectObject(self.opaque_bitmap)
        mdc2.SetBrush(wx.Brush(WindowColor))
        mdc2.SetPen(wx.TRANSPARENT_PEN)
        mdc2.DrawRectangle(0, 0, dx, dy)
        mdc = wx.MemoryDC()
        mdc.SelectObject(bitmap)
        mdc2.Blit(0, 0, dx, dy, mdc, 0, 0, useMask=True)
        mdc.SelectObject(wx.NullBitmap)
        mdc2.SelectObject(wx.NullBitmap)

        # Finally, analyze the image to find out its characteristics:
        self._analyze_bitmap()

    #-- Private Methods ------------------------------------------------------

    def _analyze_bitmap(self):
        """ Analyzes the bitmap.
        """
        # Get the image data:
        threshold = self.threshold
        bitmap = self.opaque_bitmap
        dx, dy = self.dx, self.dy
        image = bitmap.ConvertToImage()

        # Convert the bitmap data to a numpy array for analysis:
        data = reshape(fromstring(image.GetData(), uint8), (dy, dx, 3))

        # Find the horizontal slices:
        matches = []
        y, last = 0, dy - 1
        max_diff = 0.10 * dx
        while y < last:
            y_data = data[y]
            for y2 in xrange(y + 1, dy):
                if abs(y_data - data[y2]).sum() > max_diff:
                    break

            n = y2 - y
            if n >= threshold:
                matches.append((y, n))

            y = y2

        n = len(matches)
        if n == 0:
            if dy > 50:
                matches = [(0, dy)]
            else:
                matches = [(dy / 2, 1)]
        elif n > self.stretch_rows:
            matches.sort(lambda l, r: cmp(r[1], l[1]))
            matches = matches[: self.stretch_rows]

        # Calculate and save the horizontal slice sizes:
        self.fdy, self.dys = self._calculate_dxy(dy, matches)

        # Find the vertical slices:
        matches = []
        x, last = 0, dx - 1
        max_diff = 0.10 * dy
        while x < last:
            x_data = data[:, x]
            for x2 in xrange(x + 1, dx):
                if abs(x_data - data[:, x2]).sum() > max_diff:
                    break

            n = x2 - x
            if n >= threshold:
                matches.append((x, n))

            x = x2

        n = len(matches)
        if n == 0:
            if dx > 50:
                matches = [(0, dx)]
            else:
                matches = [(dx / 2, 1)]
        elif n > self.stretch_columns:
            matches.sort(lambda l, r: cmp(r[1], l[1]))
            matches = matches[: self.stretch_columns]

        # Calculate and save the vertical slice sizes:
        self.fdx, self.dxs = self._calculate_dxy(dx, matches)

        # Save the border size information:
        self.top = min(dy / 2, self.dys[0])
        self.bottom = min(dy / 2, self.dys[-1])
        self.left = min(dx / 2, self.dxs[0])
        self.right = min(dx / 2, self.dxs[-1])

        # Find the optimal size for the borders (i.e. xleft, xright, ... ):
        self._find_best_borders(data)

        # Save the background color:
        x, y = (dx / 2), (dy / 2)
        r, g, b = data[y, x]
        self.bg_color = (0x10000 * r) + (0x100 * g) + b

        # Find the best contrasting text color (black or white):
        self.content_color = self._find_best_color(data, x, y)

        # Find the best contrasting label color:
        if self.xtop >= self.xbottom:
            self.label_color = self._find_best_color(data, x, self.xtop / 2)
        else:
            self.label_color = self._find_best_color(
                data, x, dy - (self.xbottom / 2) - 1)

    def _fill(self, idc, ix, iy, idx, idy, dc, x, y, dx, dy):
        """ Performs a stretch fill of a region of an image into a region of a
            window device context.
        """
        last_x, last_y = x + dx, y + dy
        while y < last_y:
            ddy = min(idy, last_y - y)
            x0 = x
            while x0 < last_x:
                ddx = min(idx, last_x - x0)
                dc.Blit(x0, y, ddx, ddy, idc, ix, iy, useMask=True)
                x0 += ddx
            y += ddy

    def _calculate_dxy(self, d, matches):
        """ Calculate the size of all image slices for a specified set of
            matches.
        """
        if len(matches) == 1:
            d1, d2 = matches[0]

            return (d - d2, [d1, d2, d - d1 - d2])

        d1, d2 = matches[0]
        d3, d4 = matches[1]

        return (d - d2 - d4, [d1, d2, d3 - d1 - d2, d4, d - d3 - d4])

    def _find_best_borders(self, data):
        """ Find the best set of image slice border sizes (e.g. for images with
            rounded corners, there should exist a better set of borders than
            the ones computed by the image slice algorithm.
        """
        # Make sure the image size is worth bothering about:
        dx, dy = self.dx, self.dy
        if (dx < 5) or (dy < 5):
            return

        # Calculate the starting point:
        left = right = dx / 2
        top = bottom = dy / 2

        # Calculate the end points:
        last_y = dy - 1
        last_x = dx - 1

        # Mark which edges as 'scanning':
        t = b = l = r = True

        # Keep looping while at last one edge is still 'scanning':
        while l or r or t or b:

            # Calculate the current core area size:
            height = bottom - top + 1
            width = right - left + 1

            # Try to extend all edges that are still 'scanning':
            nl = (l and (left > 0) and
                  self._is_equal(data, left - 1, top, left, top, 1, height))

            nr = (r and (right < last_x) and
                  self._is_equal(data, right + 1, top, right, top, 1, height))

            nt = (t and (top > 0) and
                  self._is_equal(data, left, top - 1, left, top, width, 1))

            nb = (b and (bottom < last_y) and
                  self._is_equal(data, left, bottom + 1, left, bottom,
                                 width, 1))

            # Now check the corners of the edges:
            tl = ((not nl) or (not nt) or
                  self._is_equal(data, left - 1, top - 1, left, top, 1, 1))

            tr = ((not nr) or (not nt) or
                  self._is_equal(data, right + 1, top - 1, right, top, 1, 1))

            bl = ((not nl) or (not nb) or
                  self._is_equal(data, left - 1, bottom + 1, left, bottom,
                                 1, 1))

            br = ((not nr) or (not nb) or
                  self._is_equal(data, right + 1, bottom + 1, right, bottom,
                                 1, 1))

            # Calculate the new edge 'scanning' values:
            l = nl and tl and bl
            r = nr and tr and br
            t = nt and tl and tr
            b = nb and bl and br

            # Adjust the coordinate of an edge if it is still 'scanning':
            left -= l
            right += r
            top -= t
            bottom += b

        # Now compute the best set of image border sizes using the current set
        # and the ones we just calculated:
        self.xleft = min(self.left, left)
        self.xright = min(self.right, dx - right - 1)
        self.xtop = min(self.top, top)
        self.xbottom = min(self.bottom, dy - bottom - 1)

    def _find_best_color(self, data, x, y):
        """ Find the best contrasting text color for a specified pixel
            coordinate.
        """
        r, g, b = data[y, x]
        h, l, s = rgb_to_hls(r / 255.0, g / 255.0, b / 255.0)
        text_color = wx.BLACK
        if l < 0.50:
            text_color = wx.WHITE

        return text_color

    def _is_equal(self, data, x0, y0, x1, y1, dx, dy):
        """ Determines if two identically sized regions of an image array are
            'the same' (i.e. within some slight color variance of each other).
        """
        return (abs(data[y0: y0 + dy, x0: x0 + dx] -
                    data[y1: y1 + dy, x1: x1 + dx]).sum() < 0.10 * dx * dy)


#-------------------------------------------------------------------------
#  Returns a (possibly cached) ImageSlice:
#-------------------------------------------------------------------------

image_slice_cache = {}


def image_slice_for(image):
    """ Returns a (possibly cached) ImageSlice.
    """
    global image_slice_cache

    result = image_slice_cache.get(image)
    if result is None:
        image_slice_cache[image] = result = ImageSlice(image=image)

    return result



#!/usr/bin/env python

from fs.errors import ResourceNotFoundError
from fs.opener import opener
from fs.commands.runner import Command
import sys

class FSrm(Command):
    
    usage = """fsrm [OPTION]... [PATH]
Remove a file or directory at PATH"""
    
    def get_optparse(self):
        optparse = super(FSrm, self).get_optparse()
        optparse.add_option('-f', '--force', dest='force', action='store_true', default=False,
                            help='ignore non-existent files, never prompt')
        optparse.add_option('-i', '--interactive', dest='interactive', action='store_true', default=False,
                            help='prompt before removing')
        optparse.add_option('-r', '--recursive', dest='recursive', action='store_true', default=False,
                            help='remove directories and their contents recursively')
        return optparse
    
    def do_run(self, options, args):
                        
        interactive = options.interactive
        verbose = options.verbose
        
        for fs, path, is_dir in self.get_resources(args):                              
            if interactive:
                if is_dir:
                    msg = "remove directory '%s'?" % path
                else:
                    msg = "remove file '%s'?" % path
                if not self.ask(msg) in 'yY':
                    continue
            try:
                if is_dir:
                    fs.removedir(path, force=options.recursive)
                else:
                    fs.remove(path)
            except ResourceNotFoundError:
                if not options.force:
                    raise
            else:
                if verbose:
                    self.output("removed '%s'\n" % path)
            
        
def run():         
    return FSrm().run()
    
if __name__ == "__main__":
    sys.exit(run())    
    


""" The abstract base class for type systems. """


# Enthought library imports.
from traits.api import HasTraits


class AbstractTypeSystem(HasTraits):
    """ The abstract base class for type systems.

    A type system is responsible for:-

    1) Determining whether an object is of a particular type.
    2) Determining the MRO of a type.

    See 'PythonTypeSystem' for an implementation with standard Python
    semantics.

    """

    ###########################################################################
    # 'AbstractTypeSystem' interface.
    ###########################################################################

    def is_a(self, obj, type):
        """ Is an object an instance of the specified type? """

        raise NotImplementedError

    def get_mro(self, type):
        """ Returns the MRO of a type. """

        raise NotImplementedError

#### EOF ######################################################################



#-------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   03/02/2007
#
#-------------------------------------------------------------------------

""" Defines an editor for playing animated GIF files.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from wx.animate \
    import GIFAnimationCtrl

from traits.api \
    import Bool, Str

from traitsui.wx.editor \
    import Editor

from traitsui.basic_editor_factory \
    import BasicEditorFactory

from pyface.timer.api \
    import do_after

#-------------------------------------------------------------------------
#  '_AnimatedGIFEditor' class:
#-------------------------------------------------------------------------


class _AnimatedGIFEditor(Editor):
    """ Editor that displays an animated GIF file.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Is the animated GIF file currently playing?
    playing = Bool(True)

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self.control = GIFAnimationCtrl(parent, -1)
        self.control.GetPlayer().UseBackgroundColour(True)
        self.sync_value(self.factory.playing, 'playing', 'from')
        self.set_tooltip()

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        control = self.control
        if self.playing:
            control.Stop()

        control.LoadFile(self.value)
        self._file_loaded = True

        # Note: It seems to be necessary to Play/Stop the control to avoid a
        # hard wx crash when 'PlayNextFrame' is called the first time (must be
        # some kind of internal initialization issue):
        control.Play()
        control.Stop()

        if self.playing or self._not_first:
            self._playing_changed()
        else:
            do_after(300, self._frame_changed)

        self._not_first = True

    #-------------------------------------------------------------------------
    #  Handles the editor 'playing' trait being changed:
    #-------------------------------------------------------------------------

    def _playing_changed(self):
        """ Handles the editor 'playing' trait being changed.
        """
        if self._file_loaded:
            try:
                if self.playing:
                    self.control.Play()
                else:
                    player = self.control.GetPlayer()
                    player.SetCurrentFrame(0)
                    player.PlayNextFrame()
                    player.Stop()
            except:
                pass

#-------------------------------------------------------------------------
#  Create the editor factory object:
#-------------------------------------------------------------------------

# wxPython editor factory for animated GIF editors:


class AnimatedGIFEditor(BasicEditorFactory):

    # The editor class to be created:
    klass = _AnimatedGIFEditor

    # The optional trait used to control whether the animated GIF file is
    # playing or not:
    playing = Str



"""
__init__.py for pyugrid package

This brings in the names we want in the package.

"""

from __future__ import (absolute_import, division, print_function)

from .ugrid import UGrid
from .uvar import UVar
from .uvar import UMVar
from . import grid_io

__version__ = '0.1.8'

__all__ = ['UGrid', 'UVar', 'UMVar', 'grid_io']



from wsgiref.simple_server import make_server

from fs.osfs import OSFS
from wsgi import serve_fs
osfs = OSFS('~/')
application = serve_fs(osfs)

httpd = make_server('', 8000, application)
print "Serving on http://127.0.0.1:8000"
httpd.serve_forever()



#-------------------------------------------------------------------------------
#
#  A concrete implementation of the ITemplateDataSource interface based on the
#  implementation of the TemplateDataName class.
#
#  Write by: David C. Morrill
#
#  Date: 07/30/2007
#
#  (c) Copyright 2007 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

""" A concrete implementation of the ITemplateDataSource interface based on the
    implementation of the TemplateDataName class.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from traits.api \
    import provides

from apptools.template.template_data_name \
    import TemplateDataName

from apptools.template.itemplate_data_source \
    import ITemplateDataSource

#-------------------------------------------------------------------------------
#  'TemplateDataSource' class:
#-------------------------------------------------------------------------------

class TemplateDataSource ( TemplateDataName ):
    """ A concrete implementation of the ITemplateDataSource interface based on
        the implementation of the TemplateDataName class.
    """

    implements( ITemplateDataSource )

    #-- ITemplateDataSource Interface Implementation ---------------------------

    def name_from_data_source ( self ):
        """ Allows the object to provide a description of the possibly optional
            data binding it requires.

            Returns
            -------
            A **TemplateDataName** object describing the binding the data
            source object requires.
        """
        return self




#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: Enthought, Inc.
#
#------------------------------------------------------------------------------

""" A simple progress bar intended to run in the UI thread
"""

import wx
import time

# Enthought library imports
from traits.api import Bool, Enum, Instance, Int, Property, provides, Str

# Local imports
from pyface.i_progress_dialog import IProgressDialog, MProgressDialog
from .widget import Widget
from .window import Window

class ProgressBar(Widget):
    """ A simple progress bar dialog intended to run in the UI thread """

    #: The progress bar's parent control.
    parent = Instance(wx.Window)

    #: The progress bar's control.
    control = Instance(wx.Gauge)

    #: The orientation of the progress bar.
    direction = Enum('horizontal', 'horizontal', 'vertical')

    #: The maximum value for the progress bar.
    _max = Int

    def __init__(self, parent, minimum=0, maximum=100, direction='horizontal',
                 size=(200, -1)):
        """
        Constructs a progress bar which can be put into a panel, or optionaly,
        its own window

        """
        self._max = maximum
        self.parent = parent

        style = wx.GA_HORIZONTAL
        if direction == "vertical":
            style = wx.GA_VERTICAL

        self.control = wx.Gauge(parent, -1, maximum, style=style, size=size)

    def update(self, value):
        """ Update the progress bar to the desired value. """
        if self._max == 0:
            self.control.Pulse()
        else:
            self.control.SetValue(value)

        self.control.Update()

    def _show(self):
        # Show the parent
        self.parent.Show()

        # Show the toolkit-specific control in the parent
        self.control.Show()


class ProgressDialog(MProgressDialog, Window):
    """ A simple progress dialog window which allows itself to be updated """

    #: The progress bar
    progress_bar = Instance(ProgressBar)

    #: The window title
    title = Str

    #: The text message to display in the dialog
    message = Property()

    #: The minimum value of the progress range
    min = Int

    #: The minimum value of the progress range
    max = Int

    #: The margin around the progress bar
    margin = Int(5)

    #: Whether or not the progress dialog can be cancelled
    can_cancel = Bool(False)

    #: Whether or not to show the time taken
    show_time = Bool(False)

    #: Whether or not to show the percent completed
    show_percent = Bool(False)

    #: Whether or not the dialog was cancelled by the user
    _user_cancelled = Bool(False)

    #: The text of the message label
    _message_text = Str()

    #: The size of the dialog
    dialog_size = Instance(wx.Size)

    # Label for the 'cancel' button
    cancel_button_label = Str('Cancel')

    #: The widget showing the message text
    _message_control = Instance(wx.StaticText)

    #: The widget showing the time elapsed
    _elapsed_control = Instance(wx.StaticText)

    #: The widget showing the estimated time to completion
    _estimated_control = Instance(wx.StaticText)

    #: The widget showing the estimated time remaining
    _remaining_control = Instance(wx.StaticText)

    def __init__(self, *args, **kw):
        if 'message' in kw:
            self._message_text = kw.pop('message')

        # initialize the start time in case some tries updating
        # before open() is called
        self._start_time = 0

        super(ProgressDialog, self).__init__( *args, **kw)

    #-------------------------------------------------------------------------
    # IWindow Interface
    #-------------------------------------------------------------------------

    def open(self):
        """ Opens the window. """
        super(ProgressDialog, self).open()
        self._start_time = time.time()
        wx.GetApp().Yield(True)

    def close(self):
        """ Closes the window. """
        if self.progress_bar is not None:
            self.progress_bar.destroy()
            self.progress_bar = None

        if self._message_control is not None:
            self._message_control = None

        super(ProgressDialog, self).close()

    #-------------------------------------------------------------------------
    # IProgressDialog Interface
    #-------------------------------------------------------------------------

    def change_message(self, value):
        """ Change the displayed message in the progress dialog

        Parameters
        ----------
        message : str or unicode
            The new message to display.

        """
        self._message_text = value

        if self._message_control is not None:
            self._message_control.SetLabel(value)
            self._message_control.Update()

            msg_control_size = self._message_control.GetSize()
            self.dialog_size.x = max(self.dialog_size.x, msg_control_size.x + 2*self.margin)
            if self.control is not None:
                self.control.SetClientSize(self.dialog_size)

                self.control.GetSizer().Layout()

    def update(self, value):
        """ Update the progress bar to the desired value

        If the value is >= the maximum and the progress bar is not contained
        in another panel the parent window will be closed.

        Parameters
        ----------
        value :
            The progress value to set.
        """

        if self.progress_bar is None:
            # the developer is trying to update a progress bar which is already
            # done. Allow it, but do nothing
            return (False, False)

        self.progress_bar.update(value)

        # A bit hackish, but on Windows if another window sets focus, the
        # other window will come to the top, obscuring the progress dialog.
        # Only do this if the control is a top level window, so windows which
        # embed a progress dialog won't keep popping to the top
        # When we do embed the dialog, self.control may be None since the
        # embedding might just be grabbing the guts of the control. This happens
        # in the Traits UI ProgressEditor.

        if self.control is not None and self.control.IsTopLevel():
            self.control.Raise()

        if self.max > 0:
            percent = (float(value) - self.min)/(self.max - self.min)

            if self.show_time and (percent != 0):
                current_time = time.time()
                elapsed = current_time - self._start_time
                estimated = elapsed/percent
                remaining = estimated - elapsed

                self._set_time_label(elapsed,
                                 self._elapsed_control)
                self._set_time_label(estimated,
                                 self._estimated_control)
                self._set_time_label(remaining,
                                 self._remaining_control)

            if self.show_percent:
                self._percent_control = "%3f" % ((percent * 100) % 1)

            if value >= self.max or self._user_cancelled:
                self.close()
        else:
            if self._user_cancelled:
                self.close()

        wx.GetApp().Yield(True)

        return (not self._user_cancelled, False)

    #-------------------------------------------------------------------------
    # Private Interface
    #-------------------------------------------------------------------------

    def _on_cancel(self, event):
        self._user_cancelled = True
        self.close()

    def _on_close(self, event):
        self._user_cancelled = True
        return self.close()

    def _set_time_label(self, value, control):
        hours = value / 3600
        minutes = (value % 3600) / 60
        seconds = value % 60
        label = "%u:%02u:%02u" % (hours, minutes, seconds)

        control.SetLabel(label)

    def _get_message(self):
        return self._message_text

    def _create_buttons(self, dialog, parent_sizer):
        """ Creates the buttons. """

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        self._cancel = None

        if self.can_cancel == True:
            # 'Cancel' button.
            self._cancel = cancel = wx.Button(dialog, wx.ID_CANCEL,
                                              self.cancel_button_label)
            wx.EVT_BUTTON(dialog, wx.ID_CANCEL, self._on_cancel)
            sizer.Add(cancel, 0, wx.LEFT, 10)

            button_size = cancel.GetSize()
            self.dialog_size.x = max(self.dialog_size.x, button_size.x + 2*self.margin)
            self.dialog_size.y += button_size.y + 2*self.margin

            parent_sizer.Add(sizer, 0, wx.ALIGN_RIGHT | wx.ALL, self.margin)

    def _create_label(self, dialog, parent_sizer, text):
        local_sizer = wx.BoxSizer()
        dummy = wx.StaticText(dialog, -1, text)
        label = wx.StaticText(dialog, -1, "unknown")

        local_sizer.Add(dummy, 1, wx.ALIGN_LEFT)
        local_sizer.Add(label, 1, wx.ALIGN_LEFT | wx.ALIGN_RIGHT, self.margin)
        parent_sizer.Add(local_sizer, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.TOP, self.margin)

        return label

    def _create_gauge(self, dialog, parent_sizer):
        self.progress_bar = ProgressBar(dialog, self.min, self.max)
        parent_sizer.Add(self.progress_bar.control, 0, wx.CENTER | wx.ALL, self.margin)

        horiz_spacer = 50

        progress_bar_size = self.progress_bar.control.GetSize()
        self.dialog_size.x = max(self.dialog_size.x, progress_bar_size.x + 2*self.margin + horiz_spacer)
        self.dialog_size.y += progress_bar_size.y + 2*self.margin

    def _create_message(self, dialog, parent_sizer):
        self._message_control = wx.StaticText(dialog, -1, self.message)
        parent_sizer.Add(self._message_control, 0, wx.LEFT | wx.TOP, self.margin)

        msg_control_size = self._message_control.GetSize()
        self.dialog_size.x = max(self.dialog_size.x, msg_control_size.x + 2*self.margin)
        self.dialog_size.y += msg_control_size.y + 2*self.margin

    def _create_percent(self, dialog, parent_sizer):
        if not self.show_percent:
            return

        raise NotImplementedError

    def _create_timer(self, dialog, parent_sizer):
        if not self.show_time:
            return

        self._elapsed_control = self._create_label(dialog, parent_sizer, "Elapsed time : ")
        self._estimated_control = self._create_label(dialog, parent_sizer, "Estimated time : ")
        self._remaining_control = self._create_label(dialog, parent_sizer, "Remaining time : ")

        elapsed_size = self._elapsed_control.GetSize()
        estimated_size = self._estimated_control.GetSize()
        remaining_size = self._remaining_control.GetSize()

        timer_size = wx.Size()
        timer_size.x = max(elapsed_size.x, estimated_size.x, remaining_size.x)
        timer_size.y = elapsed_size.y + estimated_size.y + remaining_size.y

        self.dialog_size.x = max(self.dialog_size.x, timer_size.x + 2*self.margin)
        self.dialog_size.y += timer_size.y + 2*self.margin

    def _create_control(self, parent):
        """ Creates the window contents.

        This method is intended to be overridden if necessary.  By default we
        just create an empty panel.

        """
        style = wx.DEFAULT_FRAME_STYLE | wx.FRAME_NO_WINDOW_MENU | wx.CLIP_CHILDREN

        dialog = wx.Frame(parent, -1, self.title, style=style, size=self.size,
                pos=self.position)

        sizer = wx.BoxSizer(wx.VERTICAL)
        dialog.SetSizer(sizer)
        dialog.SetAutoLayout(True)
        dialog.SetBackgroundColour(wx.NullColour)

        self.dialog_size = wx.Size()

        # The 'guts' of the dialog.
        self._create_message(dialog, sizer)
        self._create_gauge(dialog, sizer)
        self._create_percent(dialog, sizer)
        self._create_timer(dialog, sizer)
        self._create_buttons(dialog, sizer)

        dialog.SetClientSize(self.dialog_size)

        dialog.CentreOnParent()

        return dialog



from svg_constants import DEFAULT_FILL, DEFAULT_STROKE
from svg_parser_utils import *

class SVGStyle(object):

    def __init__(self, inherit_from=None):
        #: The internal color
        self.fill = DEFAULT_FILL

        #: How to determine the "insideness" of a point. Possible values are
        #: "nonzero" and "evenodd". See mozilla documentation for more details:
        #: http://www.w3.org/TR/SVG/painting.html
        self.fill_rule = 'nonzero'

        #: The color of the surrounding outline
        self.stroke = DEFAULT_STROKE

        #: The width of the surrounding outline
        self.stroke_width = 1

        #: Overall opacity (multiplied by other elements)
        self.opacity = 1.0

        #: Controls the pattern of dashes and gaps used to stroke path. List
        #: of alternating dashes and gaps.
        self.stroke_dasharray = []

        #: The maximum ratio of the distance between a line joints inner
        #: connection and outer miter edge vs the line width
        self.stroke_miterlimit = 4

        #: The line cap, possible values are 'butt', 'square', 'round'
        self.stroke_linecap = 'butt'

        #: Current style font family
        self.font_family = 'Arial'

        #: Current font size
        self.font_size = 8

        if inherit_from:
            self.fill = inherit_from.fill
            self.stroke = inherit_from.stroke
            self.fill_rule = inherit_from.fill_rule
            self.stroke_width = inherit_from.stroke_width
            self.font_family = inherit_from.font_family
            self.font_size = inherit_from.font_size

    def from_element(self, element):
        """Read relevant attributes off XML element"""
        self.fill = parse_color(element.get('fill'), self.fill)
        self.fill_rule = element.get('fill-rule', self.fill_rule)

        self.stroke = parse_color(element.get('stroke'), self.stroke)
        self.stroke_width = parse_float(element.get('stroke-width', '1.0'))

        self.opacity *= float(element.get('opacity', 1))
        self.fill_opacity = float(element.get('fill-opacity', 1))
        self.stroke_opacity = float(element.get('stroke-opacity', 1))
        self.stroke_linejoin = element.get('stroke-linejoin', 'miter')
        self.stroke_miterlimit = float(element.get('stroke-miterlimit', 4))
        self.stroke_linecap = element.get('stroke-linecap', 'butt')

        self.font_family = element.get('font-family', 'Arial')
        self.font_size = parse_float(element.get('font-size', '8'))

        dash_array = element.get('stroke-dasharray', None)
        if dash_array:
            self.stroke_dasharray = [float(x.strip()) for x in dash_array.split(',')]

        style = element.get('style')
        if style:
            style_dict = parse_style(style)
            if 'font-family' in style_dict:
                self.font_family = style_dict['font-family']
            if 'font-size' in style_dict:
                self.font_size = parse_float(style_dict['font-size'])
            if 'fill' in style_dict:
                self.fill = parse_color(style_dict['fill'])
            if 'fill-opacity' in style_dict:
                self.fill_opacity *= float(style_dict['fill-opacity'])
            if 'stroke' in style_dict:
                self.stroke = parse_color(style_dict['stroke'])
            if 'stroke-opacity' in style_dict:
                self.stroke_opacity *= float(style_dict['stroke-opacity'])
            if 'stroke-width' in style_dict:
                sw = style_dict['stroke-width']
                self.stroke_width = parse_float(sw)
            if 'stroke-dasharray' in style_dict:
                dash_array = style_dict['stroke-dasharray']
                if dash_array and dash_array != 'none':
                    self.stroke_dasharray = [float(x.strip()) for x in dash_array.split(',')]
            if 'stroke-linejoin' in style_dict:
                self.stroke_linejoin = style_dict['stroke-linejoin']
            if 'stroke-linecap' in style_dict:
                self.stroke_linecap = style_dict['stroke-linecap']
            if 'opacity' in style_dict:
                self.fill_opacity *= float(style_dict['opacity'])
                self.stroke_opacity *= float(style_dict['opacity'])
            if 'fill-rule' in style_dict:
                self.fill_rule = style_dict['fill-rule']
        if isinstance(self.stroke, list):
            self.stroke[3] = int(self.opacity * self.stroke_opacity * self.stroke[3])
        if isinstance(self.fill, list):
            self.fill[3] = int(self.opacity * self.fill_opacity * self.fill[3])

    def parse_style_attribute(self, attr):
        pass

    def parse_fill_attribute(self, attr):
        pass

    def parse_fillrule(self, attr):
        pass

    def parse_stroke_attribute(self, attr):
        pass

    def parse_stroke_width(self, attr):
        pass

    def parse_stroke_dasharray(self, attr):
        pass


#-------------------------------------------------------------------------------
#
#  Defines the Template class, which provides a default implementation of the
#  ITemplate interface that can be used when defining templatizable objects.
#
#  Written by: David C. Morrill
#
#  Date: 07/26/2007
#
#  (c) Copyright 2007 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

""" Defines the Template class, which provides a default implementation of the
    ITemplate interface that can be used when defining templatizable objects.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from traits.api \
    import HasPrivateTraits, Undefined, provides

from itemplate \
    import ITemplate

#-------------------------------------------------------------------------------
#  Metadata filter for extracting 'object' and 'datasource' items:
#-------------------------------------------------------------------------------

def instance_or_datasource ( value ):
    return value in ( 'instance', 'datasource' )

#-------------------------------------------------------------------------------
#  Metadata filter for extracting traits with no 'template' metadata:
#-------------------------------------------------------------------------------

def is_none ( value ):
    return (value is None)

#-------------------------------------------------------------------------------
#  'Template' class:
#-------------------------------------------------------------------------------

class Template ( HasPrivateTraits ):
    """ Defines the Template class, which provides a default implementation of
        the ITemplate interface that can be used when defining templatizable
        objects.
    """

    implements( ITemplate )

    #-- 'ITemplate' Interface Implementation -----------------------------------

    def names_from_template ( self ):
        """ Returns a list of **TemplateDataName** objects, one for each
            bindable data source contained within the template. Each
            **TemplateDataName** object supports unresolved bindings in
            addition to resolved bindings. Also, a **TemplateDataName** may be
            marked as *optional*, meaning the if the name remains unresolved,
            the application can take an alternative action (e.g. omit the data
            from a plot, or substitute default data, etc.).

            Returns
            -------
            A list of **TemplateDataName** objects, one for each bindable data
            source contained in the template.
        """
        data_names = []

        # Recursively propagate the request to all contained templatizable
        # objects:
        for value in self.get( template = 'instance' ).values():
            if isinstance( value, list ):
                for v in value:
                    data_names.extend( v.names_from_template() )
            elif value is not None:
                data_names.extend( value.names_from_template() )

        # Add the bindings for all of our local data source objects:
        for value in self.get( template = 'datasource' ).values():
            if isinstance( value, list ):
                for v in value:
                    data_names.append( v.name_from_data_source() )
            elif value is not None:
                data_names.append( value.name_from_data_source() )

        # Return the list of TemplateDataName objects we collected:
        return data_names

    def object_from_template ( self ):
        """ Activates the object from its template data.

            Returns
            -------
            The original object.
        """
        # Convert all our data source objects to 'live' objects:
        for value in self.get( template = 'datasource' ).values():
            if isinstance( value, list ):
                for v in value:
                    v.object_from_template()
            elif value is not None:
                value.object_from_template()

        # Recursively propagate the request to all contained templatizable
        # objects:
        for value in self.get( template = 'instance' ).values():
            if isinstance( value, list ):
                for v in value:
                    v.object_from_template()
            elif value is not None:
                value.object_from_template()

        # Reset all of our 'live' objects to undefined:
        for name in self.trait_names( template = 'derived' ):
            setattr( self, name, Undefined )

        # Return ourselves as the new, live version of the object:
        return self

    def template_from_object ( self ):
        """ Returns a *templatized* version of the object that is safe for
            serialization.

            Returns
            -------
            A new object of the same class as the original.
        """
        # Mark all traits without 'template' metadata as 'transient':
        for trait in self.traits( template = is_none ).values():
            trait.transient = True

        # Collect all of the data source and templatizable objects we contain:
        contents = self.get( template = instance_or_datasource )

        # Convert them all to new objects in template form:
        for name, value in contents.items():
            if isinstance( value, list ):
                contents[ name ] = [ v.template_from_object() for v in value ]
            else:
                contents[ name ] = value.template_from_object()

        # Add all of the simple values which can just be copied:
        contents.update( self.get( template = 'copy' ) )

        # Finally return a new instance of our class containing just the
        # templatizable values:
        return self.__class__( **contents )

    def activate_template ( self ):
        """ Converts all contained 'TDerived' objects to real objects using the
            template traits of the object. This method must be overridden in
            subclasses.

            Returns
            -------
            None
        """
        raise NotImplementedError




"""Sample panes for HexEdit

"""
# Enthought library imports.
from pyface.tasks.api import TaskLayout, PaneItem, HSplitter, VSplitter

import panes


# The project ID must be changed when the pane layout changes, otherwise
# the new panes won't be displayed because the previous saved state of the
# application will be loaded.  Changing the project ID forces the framework to
# honor the new layout, because there won't be a saved state of this new ID.

# The saved state is stored in ~/.config/Omnivore/tasks/wx/application_memento

# Removing this file will cause the default layout to be used.  The saved state
# is only updated when quitting the application; if the application is killed
# (or crashes!) the saved state is not updated.

pane_layout_version = 'v2'


def pane_layout():
    """ Create the default task layout, which is overridded by the user's save
    state if it exists.
    """
    return TaskLayout()


def pane_create():
    """ Create all the pane objects available for the task (regardless
    of visibility -- visibility is handled in the task activation method
    MaproomTask.activated)
    """
    return []



#-------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   07/09/2007
#
#-------------------------------------------------------------------------

""" Defines a ThemedWindow base class for creating themed windows.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

from traits.api \
    import HasPrivateTraits, HasTraits, Instance, Str, Enum, Any, Bool

from traitsui.ui_traits \
    import ATheme

from helper \
    import init_wx_handlers, BufferDC

#-------------------------------------------------------------------------
#  'ThemedWindow' class:
#-------------------------------------------------------------------------


class ThemedWindow(HasPrivateTraits):

    #-- Public Traits --------------------------------------------------------

    # The theme associated with this window:
    theme = ATheme

    # The default alignment to use:
    default_alignment = Enum('left', 'center', 'right')

    # The current mouse event state:
    state = Str('normal')

    # Optional controller used for overriding event handling:
    controller = Instance(HasTraits)

    # Should debugging information be overlaid on the theme?
    debug = Bool(False)

    #-- Public Methods -------------------------------------------------------

    def init_control(self):
        """ Initializes the underlying wx.Window object.
        """
        init_wx_handlers(self.control, self)

    def in_control(self, x, y):
        """ Returns whether a specified (x,y) coordinate is inside the control
            or not.
        """
        wdx, wdy = self.control.GetClientSize()
        return ((0 <= x < wdx) and (0 <= y < wdy))

    def refresh(self):
        """ Refreshes the contents of the control.
        """
        if self.control is not None:
            self.control.Refresh()

    def capture_mouse(self):
        """ Grab control of the mouse and indicate that we are controlling it.
        """
        if not self.control.HasCapture():
            self.control.CaptureMouse()

    def release_mouse(self):
        """ Grab control of the mouse and indicate that we are controlling it.
        """
        if self.control.HasCapture():
            self.control.ReleaseMouse()

    #-- Trait Event Handlers -------------------------------------------------

    def _theme_changed(self):
        """ Handles the 'theme' trait being changed.
        """
        self.refresh()

    #-- wxPython Event Handlers ----------------------------------------------

    def _erase_background(self, event):
        """ Do not erase the background here (do it in the 'on_paint' handler).
        """
        pass

    def _paint(self, event):
        """ Paint the background using the associated ImageSlice object.
        """
        dc = BufferDC(self.control)
        self._paint_bg(dc)
        self._paint_fg(dc)
        dc.copy()

    def _paint_bg(self, dc):
        """ Paints the background into the supplied device context using the
            associated ImageSlice object and returns the image slice used (if
            any).
        """
        from image_slice import paint_parent

        # Repaint the parent's theme (if necessary):
        paint_parent(dc, self.control)

        # Draw the background theme (if any):
        if self.theme is not None:
            slice = self.theme.image_slice
            if slice is not None:
                wdx, wdy = self.control.GetClientSize()
                slice.fill(dc, 0, 0, wdx, wdy, True)

                if self.debug:
                    dc.SetPen(wx.Pen(wx.RED))
                    dc.SetBrush(wx.TRANSPARENT_BRUSH)
                    theme = self.theme
                    border = theme.border
                    dc.DrawRectangle(border.left, border.top,
                                     wdx - border.right - border.left,
                                     wdy - border.bottom - border.top)
                    dc.DrawRectangle(border.left + 3, border.top + 3,
                                     wdx - border.right - border.left - 6,
                                     wdy - border.bottom - border.top - 6)
                    content = theme.content
                    x = slice.xleft + content.left
                    y = slice.xtop + content.top
                    dc.DrawRectangle(
                        x -
                        1,
                        y -
                        1,
                        wdx -
                        slice.xright -
                        content.right -
                        x +
                        2,
                        wdy -
                        slice.xbottom -
                        content.bottom -
                        y +
                        2)

                    label = theme.label
                    if slice.xtop >= slice.xbottom:
                        y, dy = 0, slice.xtop
                    else:
                        y, dy = wdy - slice.xbottom, slice.xbottom

                    if dy >= 13:
                        x = slice.xleft + label.left
                        y += label.top
                        dc.DrawRectangle(
                            x -
                            1,
                            y -
                            1,
                            wdx -
                            slice.xright -
                            label.right -
                            x +
                            2,
                            dy -
                            label.bottom -
                            label.top +
                            2)

    def _paint_fg(self, dc):
        """ Paints the foreground of the window into the supplied device
            context.
        """
        pass

    def _size(self, event):
        """ Handles the control being resized.
        """
        control = self.control
        if control is not None:
            control.Layout()
            control.Refresh()

    def _left_down(self, event):
        """ Handles a left mouse button down event.
        """
        self.control.SetFocus()
        self.capture_mouse()
        self._mouse_event('left_down', event)

    def _left_up(self, event):
        """ Handles a left mouse button up event.
        """
        self.release_mouse()
        self._mouse_event('left_up', event)

    def _left_dclick(self, event):
        """ Handles a left mouse button double click event.
        """
        self.capture_mouse()
        self._mouse_event('left_dclick', event)

    def _middle_down(self, event):
        """ Handles a middle mouse button down event.
        """
        self.capture_mouse()
        self._mouse_event('middle_down', event)

    def _middle_up(self, event):
        """ Handles a middle mouse button up event.
        """
        self.release_mouse()
        self._mouse_event('middle_up', event)

    def _middle_dclick(self, event):
        """ Handles a middle mouse button double click event.
        """
        self.capture_mouse()
        self._mouse_event('middle_dclick', event)

    def _right_down(self, event):
        """ Handles a right mouse button down event.
        """
        self.capture_mouse()
        self._mouse_event('right_down', event)

    def _right_up(self, event):
        """ Handles a right mouse button up event.
        """
        self.release_mouse()
        self._mouse_event('right_up', event)

    def _right_dclick(self, event):
        """ Handles a right mouse button double click event.
        """
        self.capture_mouse()
        self._mouse_event('right_dclick', event)

    def _motion(self, event):
        """ Handles a mouse move event.
        """
        self._mouse_event('motion', event)

    def _enter(self, event):
        """ Handles the mouse entering the window event.
        """
        self._mouse_event('enter', event)

    def _leave(self, event):
        """ Handles the mouse leaving the window event.
        """
        self._mouse_event('leave', event)

    def _wheel(self, event):
        """ Handles a mouse wheel event.
        """
        self._mouse_event('wheel', event)

    #-- Private Methods ------------------------------------------------------

    def _mouse_event(self, name, event):
        """ Routes a mouse event to the proper handler (if any).
        """
        method_name = '%s_%s' % (self.state, name)
        method = None

        if self.controller is not None:
            method = getattr(self.controller, method_name, None)

        if method is None:
            method = getattr(self, method_name, None)

        if method is not None:
            method(event.GetX(), event.GetY(), event)



"""
fs.contrib.archivefs
========

A FS object that represents the contents of an archive.

"""

import time
import stat
import datetime
import os.path

from fs.base import *
from fs.path import *
from fs.errors import *
from fs.filelike import StringIO
from fs import mountfs

import libarchive

ENCODING = libarchive.ENCODING


class SizeUpdater(object):
    '''A file-like object to allow writing to a file within the archive. When closed
    this object will update the archive entry's size within the archive.'''
    def __init__(self, entry, stream):
        self.entry = entry
        self.stream = stream
        self.size = 0

    def __del__(self):
        self.close()

    def write(self, data):
        self.size += len(data)
        self.stream.write(data)

    def close(self):
        self.stream.close()
        self.entry.size = self.size


class ArchiveFS(FS):
    """A FileSystem that represents an archive supported by libarchive."""

    _meta = { 'thread_safe' : True,
              'virtual' : False,
              'read_only' : False,
              'unicode_paths' : True,
              'case_insensitive_paths' : False,
              'network' : False,
              'atomic.setcontents' : False
             }

    def __init__(self, f, mode='r', format=None, thread_synchronize=True):
        """Create a FS that maps on to an archive file.

        :param f: a (system) path, or a file-like object
        :param format: required for 'w' mode. The archive format ('zip, 'tar', etc)
        :param thread_synchronize: set to True (default) to enable thread-safety
        """
        super(ArchiveFS, self).__init__(thread_synchronize=thread_synchronize)
        if isinstance(f, basestring):
            self.fileobj = None
            self.root_path = f
        else:
            self.fileobj = f
            self.root_path = getattr(f, 'name', None)
        self.contents = PathMap()
        self.archive = libarchive.SeekableArchive(f, format=format, mode=mode)
        if 'r' in mode:
            for item in self.archive:
                for part in recursepath(item.pathname)[1:]:
                    part = relpath(part)
                    if part == item.pathname:
                        self.contents[part] = item
                    else:
                        self.contents[part] = libarchive.Entry(pathname=part, mode=stat.S_IFDIR, size=0, mtime=item.mtime)

    def __str__(self):
        return "<ArchiveFS: %s>" % self.root_path

    def __unicode__(self):
        return u"<ArchiveFS: %s>" % self.root_path

    def getmeta(self, meta_name, default=NoDefaultMeta):
        if meta_name == 'read_only':
            return self.read_only
        return super(ArchiveFS, self).getmeta(meta_name, default)

    @synchronize
    def close(self):
        if getattr(self, 'archive', None) is None:
            return
        self.archive.close()

    @synchronize
    def open(self, path, mode="r", **kwargs):
        path = normpath(relpath(path))
        if path == '':
            # We need to open the archive itself, not one of it's entries.
            return file(self.root_path, mode)
        if 'a' in mode:
            raise Exception('Unsupported mode ' + mode)
        if 'r' in mode:
            return self.archive.readstream(path)
        else:
            entry = self.archive.entry_class(pathname=path, mode=stat.S_IFREG, size=0, mtime=time.time())
            self.contents[path] = entry
            return SizeUpdater(entry, self.archive.writestream(path))

    @synchronize
    def getcontents(self, path, mode="rb", encoding=None, errors=None, newline=None):
        if not self.exists(path):
            raise ResourceNotFoundError(path)
        with self.open(path, mode, encoding=encoding, errors=errors, newline=newline) as f:
            return f.read()

    def desc(self, path):
        return "%s in zip file" % path

    def getsyspath(self, path, allow_none=False):
        path = normpath(path).lstrip('/')
        return join(self.root_path, path)

    def isdir(self, path):
        info = self.getinfo(path)
        # Don't use stat.S_ISDIR, it won't work when mode == S_IFREG | S_IFDIR.
        return info.get('st_mode', 0) & stat.S_IFDIR == stat.S_IFDIR

    def isfile(self, path):
        info = self.getinfo(path)
        # Don't use stat.S_ISREG, it won't work when mode == S_IFREG | S_IFDIR.
        return info.get('st_mode', 0) & stat.S_IFREG == stat.S_IFREG

    def exists(self, path):
        path = normpath(path).lstrip('/')
        if path == '':
            # We are being asked about root (the archive itself)
            return True
        return path in self.contents

    def listdir(self, path="/", wildcard=None, full=False, absolute=False, dirs_only=False, files_only=False):
        return self._listdir_helper(path, self.contents.names(path), wildcard, full, absolute, dirs_only, files_only)

    def makedir(self, dirname, recursive=False, allow_recreate=False):
        entry = self.archive.entry_class(pathname=dirname, mode=stat.S_IFDIR, size=0, mtime=time.time())
        self.contents[dirname] = entry
        self.archive.write(entry)

    @synchronize
    def getinfo(self, path):
        if not self.exists(path):
            raise ResourceNotFoundError(path)
        path = normpath(path).lstrip('/')
        info = { 'size': 0 }
        entry = self.contents.get(path)
        for attr in dir(entry):
            if attr.startswith('_'):
                continue
            elif attr == 'mtime':
                info['modified_time'] = datetime.datetime.fromtimestamp(entry.mtime)
            elif attr == 'mode':
                info['st_mode'] = entry.mode
            else:
                value = getattr(entry, attr)
                if callable(value):
                    continue
                info[attr] = value
        return info

    def getsize(self, path):
        return self.getinfo(path)['st_size']


class ArchiveMountFS(mountfs.MountFS):
    '''A subclass of MountFS that automatically identifies archives. Once identified
    archives are mounted in place of the archive file.'''

    def __init__(self, rootfs, auto_close=True, auto_mount=True, max_size=None):
        self.auto_mount = auto_mount
        self.max_size = max_size
        super(ArchiveMountFS, self).__init__(auto_close=auto_close)
        self.rootfs = rootfs
        self.mountdir('/', rootfs)

    @synchronize
    def close(self):
        # Close and delete references to any other fs instances.
        if self.rootfs is not None:
            self.rootfs.close()
            self.rootfs = None
        super(ArchiveMountFS, self).close()

    def ismount(self, path):
        "Checks if the given path has a file system mounted on it."
        try:
            object = self.mount_tree[path]
        except KeyError:
            return False
        return isinstance(object, mountfs.MountFS.DirMount)

    def unmount(self, path):
        """Unmounts a path.

        :param path: Path to unmount

        """
        # This might raise a KeyError, but that is what MountFS will do, so
        # shall we.
        fs = self.mount_tree.pop(path)
        # TODO: it may be necessary to remember what paths were auto-mounted,
        # so we can close those here. It may not be safe to close a file system
        # that the user provided. However, it is definitely NOT safe to leave
        # one open.
        if callable(getattr(fs, 'close', None)):
            fs.close()

    def _delegate(self, path, auto_mount=True):
        """A _delegate() override that will automatically mount archives that are
        encountered in the path. For example, the path /foo/bar.zip/baz.txt contains
        the archive path /foo/bar.zip. If this archive can be mounted by ArchiveFS,
        it will be. Then the file system call will be delegated to that mounted file
        system, which will act upon /baz.txt within the archive. This is lazy
        initialization which means users of this class need not crawl the file system
        for archives and mount them all up-front.

        This behavior can be overridden by self.auto_mount=False or by passing the
        auto_mount=False keyword argument.
        """
        if self.auto_mount and auto_mount:
            for ppath in recursepath(path)[1:]:
                if self.ismount(ppath):
                    # If something is already mounted here, no need to continue.
                    break
                if libarchive.is_archive_name(ppath):
                    # It looks like an archive, we might mount it.
                    # First check that the size is acceptable.
                    if self.max_size:
                        if self.rootfs.exists(ppath) and \
                           self.rootfs.getsize(ppath) > self.max_size:
                            break
                    # Looks good, the proof is in the pudding, so let's try to
                    # mount this *supposed* archive...
                    full_path = self.rootfs.getsyspath(ppath)
                    try:
                        # TODO: it would be really nice if we could open the path using
                        # self.rootfs.open(), that way we could support archives on a file
                        # system other than osfs (even nested archives). However, the libarchive
                        # wrapper is not sophisticated enough to handle a Python file-like object,
                        # it uses an actual fd.
                        self.mountdir(ppath, ArchiveFS(full_path, 'r'))
                        # That worked!! Stop recursing path, we support just one archive per path!
                        break
                    except:
                        # Must NOT have been an archive after all, but maybe
                        # there is one deeper in the directory...
                        continue
        return super(ArchiveMountFS, self)._delegate(path)

    def getsyspath(self, path, allow_none=False):
        """A getsyspath() override that returns paths relative to the root fs."""
        root = self.rootfs.getsyspath('/', allow_none=allow_none)
        if root:
            return join(root, path.lstrip('/'))

    def open(self, path, *args, **kwargs):
        """An open() override that opens an archive. It is not fooled by mounted
        archives. If the path is a mounted archive, it is unmounted and the archive
        file is opened and returned."""
        if libarchive.is_archive_name(path) and self.ismount(path):
            self.unmount(path)
        fs, _mount_path, delegate_path = self._delegate(path, auto_mount=False)
        return fs.open(delegate_path, *args, **kwargs)

    def getinfo(self, path):
        """A getinfo() override that allows archives to masqueraded as directories.
        If the path is not an archive, the call is delegated. In the event that the
        path is an archive, that archive is mounted to ensure it can actually be
        treaded like a directory."""
        fs, _mount_path, delegate_path = self._delegate(path)
        if isinstance(fs, ArchiveFS) and path == _mount_path:
            info = self.rootfs.getinfo(path)
            info['st_mode'] = info.get('st_mode', 0) | stat.S_IFDIR
            return info
        return super(ArchiveMountFS, self).getinfo(path)

    def isdir(self, path):
        """An isdir() override that allows archives to masquerade as directories. If
        the path is not an archive, the call is delegated. In the event that the path
        is an archive, that archive is mounted to ensure it can actually be treated
        like a directory."""
        fs, _mount_path, delegate_path = self._delegate(path)
        if isinstance(fs, ArchiveFS) and path == _mount_path:
            # If the path is an archive mount point, it is a directory.
            return True
        return super(ArchiveMountFS, self).isdir(path)

    def isfile(self, path):
        """An isfile() override that checks if the given path is a file or not. It is
        not fooled by a mounted archive. If the path is not an archive, True is returned.
        If the path is not an archive, the call is delegated."""
        fs, _mount_path, delegate_path = self._delegate(path)
        if isinstance(fs, ArchiveFS) and path == _mount_path:
            # If the path is an archive mount point, it is a file.
            return True
        else:
            return fs.isfile(delegate_path)

    def getsize(self, path):
        """A getsize() override that returns the size of an archive. It is not fooled by
        a mounted archive. If the path is not an archive, the call is delegated."""
        fs, _mount_path, delegate_path = self._delegate(path, auto_mount=False)
        if isinstance(fs, ArchiveFS) and path == _mount_path:
            return self.rootfs.getsize(path)
        else:
            return fs.getsize(delegate_path)

    def remove(self, path):
        """A remove() override that deletes an archive directly. It is not fooled
        by a mounted archive. If the path is not an archive, the call is delegated."""
        if libarchive.is_archive_name(path) and self.ismount(path):
            self.unmount(path)
        fs, _mount_path, delegate_path = self._delegate(path, auto_mount=False)
        return fs.remove(delegate_path)

    def makedir(self, path, *args, **kwargs):
        """A makedir() override that handles creation of a directory at an archive
        location properly. If the path is not an archive, the call is delegated."""
        # If the caller is trying to create a directory where an archive lives
        # we should raise an error. In the case when allow_recreate=True, this
        # call would succeed without the check below.
        fs, _mount_path, delegate_path = self._delegate(path, auto_mount=False)
        if isinstance(fs, ArchiveFS) and path == _mount_path:
            raise ResourceInvalidError(path, msg="Cannot create directory, there's "
                                       "already a file of that name: %(path)s")
        return fs.makedir(delegate_path, *args, **kwargs)

    def copy(self, src, dst, overwrite=False, chunk_size=1024*64):
        """An optimized copy() that will skip mounting an archive if one is involved
        as either the src or dst. This allows the file containing the archive to be
        copied."""
        src_is_archive = libarchive.is_archive_name(src)
        # If src path is a mounted archive, unmount it.
        if src_is_archive and self.ismount(src):
            self.unmount(src)
        # Now delegate the path, if the path is an archive, don't remount it.
        srcfs, _ignored, src = self._delegate(src, auto_mount=(not src_is_archive))
        # Follow the same steps for dst.
        dst_is_archive = libarchive.is_archive_name(dst)
        if dst_is_archive and self.ismount(dst):
            self.unmount(dst)
        dstfs, _ignored, dst = self._delegate(dst, auto_mount=(not dst_is_archive))
        # srcfs, src and dstfs, dst are now the file system and path for our src and dst.
        if srcfs is dstfs and srcfs is not self:
            # Both src and dst are on the same fs, let it do the copy.
            srcfs.copy(src, dst, overwrite=overwrite, chunk_size=chunk_size)
        else:
            # Src and dst are on different file systems. Just do the copy...
            srcfd = None
            try:
                srcfd = srcfs.open(src, 'rb')
                dstfs.setcontents(dst, srcfd, chunk_size=chunk_size)
            except ResourceNotFoundError:
                if srcfs.exists(src) and not dstfs.exists(dirname(dst)):
                    raise ParentDirectoryMissingError(dst)
            finally:
                if srcfd:
                    srcfd.close()

    def move(self, src, dst, overwrite=False, chunk_size=1024*64):
        """An optimized move() that delegates the work to the overridden copy() and
        remove() methods."""
        self.copy(src, dst, overwrite=overwrite, chunk_size=chunk_size)
        self.remove(src)

    def rename(self, src, dst):
        """An rename() implementation that ensures the rename does not span
        file systems. It also ensures that an archive can be renamed (without
        trying to mount either the src or destination paths)."""
        src_is_archive = libarchive.is_archive_name(src)
        # If src path is a mounted archive, unmount it.
        if src_is_archive and self.ismount(src):
            self.unmount(src)
        # Now delegate the path, if the path is an archive, don't remount it.
        srcfs, _ignored, src = self._delegate(src, auto_mount=(not src_is_archive))
        # Follow the same steps for dst.
        dst_is_archive = libarchive.is_archive_name(dst)
        if dst_is_archive and self.ismount(dst):
            self.unmount(dst)
        dstfs, _ignored, dst = self._delegate(dst, auto_mount=(not dst_is_archive))
        # srcfs, src and dstfs, dst are now the file system and path for our src and dst.
        if srcfs is dstfs and srcfs is not self:
            # Both src and dst are on the same fs, let it do the copy.
            return srcfs.rename(src, dst)
        raise OperationFailedError("rename resource", path=src)

    def walk(self,
             path="/",
             wildcard=None,
             dir_wildcard=None,
             search="breadth",
             ignore_errors=False,
             archives_as_files=True):
        """Walks a directory tree and yields the root path and contents.
        Yields a tuple of the path of each directory and a list of its file
        contents.

        :param path: root path to start walking
        :type path: string
        :param wildcard: if given, only return files that match this wildcard
        :type wildcard: a string containing a wildcard (e.g. `*.txt`) or a callable that takes the file path and returns a boolean
        :param dir_wildcard: if given, only walk directories that match the wildcard
        :type dir_wildcard: a string containing a wildcard (e.g. `*.txt`) or a callable that takes the directory name and returns a boolean
        :param search: a string identifying the method used to walk the directories. There are two such methods:

             * ``"breadth"`` yields paths in the top directories first
             * ``"depth"`` yields the deepest paths first

        :param ignore_errors: ignore any errors reading the directory
        :type ignore_errors: bool
        :param archives_as_files: treats archives as files rather than directories.
        :type ignore_errors: bool

        :rtype: iterator of (current_path, paths)

        """
        path = normpath(path)

        def isdir(path):
            if not self.isfile(path):
                return True
            if not archives_as_files and self.ismount(path):
                return True
            return False

        def listdir(path, *args, **kwargs):
            dirs_only = kwargs.pop('dirs_only', False)
            if ignore_errors:
                try:
                    listing = self.listdir(path, *args, **kwargs)
                except:
                    return []
            else:
                listing = self.listdir(path, *args, **kwargs)
            if dirs_only:
                listing = filter(isdir, listing)
            return listing

        if wildcard is None:
            wildcard = lambda f:True
        elif not callable(wildcard):
            wildcard_re = re.compile(fnmatch.translate(wildcard))
            wildcard = lambda fn:bool (wildcard_re.match(fn))

        if dir_wildcard is None:
            dir_wildcard = lambda f:True
        elif not callable(dir_wildcard):
            dir_wildcard_re = re.compile(fnmatch.translate(dir_wildcard))
            dir_wildcard = lambda fn:bool (dir_wildcard_re.match(fn))

        if search == "breadth":

            dirs = [path]
            dirs_append = dirs.append
            dirs_pop = dirs.pop
            while dirs:
                current_path = dirs_pop()
                paths = []
                paths_append = paths.append
                try:
                    for filename in listdir(current_path):
                        path = pathcombine(current_path, filename)
                        if isdir(path):
                            if dir_wildcard(path):
                                dirs_append(path)
                        else:
                            if wildcard(filename):
                                paths_append(filename)
                except ResourceNotFoundError:
                    # Could happen if another thread / process deletes something whilst we are walking
                    pass

                yield (current_path, paths)

        elif search == "depth":

            def recurse(recurse_path):
                try:
                    for path in listdir(recurse_path, wildcard=dir_wildcard, full=True, dirs_only=True):
                        for p in recurse(path):
                            yield p
                except ResourceNotFoundError:
                    # Could happen if another thread / process deletes something whilst we are walking
                    pass
                yield (recurse_path, listdir(recurse_path, wildcard=wildcard, files_only=True))

            for p in recurse(path):
                yield p

        else:
            raise ValueError("Search should be 'breadth' or 'depth'")


def main():
    ArchiveFS()


if __name__ == '__main__':
    main()




#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/13/2004
#
#------------------------------------------------------------------------------

""" Defines the UIInfo class used to represent the object and editor content of
    an active Traits-based user interface.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import HasPrivateTraits, Instance, Constant, Bool

#-------------------------------------------------------------------------
#  'UIInfo' class:
#-------------------------------------------------------------------------


class UIInfo(HasPrivateTraits):
    """ Represents the object and editor content of an active Traits-based
    user interface
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Bound to a UI object at UIInfo construction time
    ui = Instance('traitsui.ui.UI', allow_none=True)

    # Indicates whether the UI has finished initialization
    initialized = Bool(False)

    #-------------------------------------------------------------------------
    #  Bind's all of the associated context objects as traits of the object:
    #-------------------------------------------------------------------------

    def bind_context(self):
        """ Binds all of the associated context objects as traits of the
            object.
        """
        for name, value in self.ui.context.items():
            self.bind(name, value)

    #-------------------------------------------------------------------------
    #  Binds a name to a value if it is not already bound:
    #-------------------------------------------------------------------------

    def bind(self, name, value, id=None):
        """ Binds a name to a value if it is not already bound.
        """
        if id is None:
            id = name

        if not hasattr(self, name):
            self.add_trait(name, Constant(value))
            if id != '':
                self.ui._names.append(id)
